"""

IP related tools.

"""

import sys
import socket
import fcntl
import struct
import array

from .compat import PY3

# Maximum number of interfaces to return
MAX_INTERFACES = 100


def get_all_interfaces():
    """
    Returns a list of tuples, giving the network name and ipv4 address
    [('lo', '127.0.0.1'), ('eth0', '10.0.2.15'), ('eth1', '192.168.33.10')]

    Uses https://linux.die.net/man/7/netdevice system call

    May raise an IOError if platform is unsupported
    """
    _is_64bits = sys.maxsize > 2 ** 32
    struct_size = 40 if _is_64bits else 32

    if_buffer = array.array("B", b"\0" * MAX_INTERFACES * struct_size)
    if_pointer, buffer_size = if_buffer.buffer_info()

    _socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        fh = _socket.fileno()
        ifaces_bytes = fcntl.ioctl(
            fh, 0x8912, struct.pack("iL", buffer_size, if_pointer)  # SIOCGIFCONF
        )
        if_buffer_size, _pointer = struct.unpack("iL", ifaces_bytes)

    finally:
        _socket.close()

    ifaces = if_buffer.tostring()
    interfaces = []
    for offset in range(0, if_buffer_size, struct_size):
        _iface = ifaces[offset : offset + struct_size]
        name, _, _ = _iface.partition(b"\0")
        ip = socket.inet_ntoa(_iface[20:24])  # struct sockaddr ifr_hwaddr;
        if not PY3:
            ip = ip.decode("utf-8")
        interfaces.append((name.decode("ascii", "replace"), ip))

    return interfaces


if __name__ == "__main__":
    # python -m dataplicity.iptools
    for name, ip in get_all_interfaces():
        print("{}\t{}".format(name, ip))



"""
On RPi, threading.Event objects block indefinitely when the clock
changes, causing agent to hang. This thread kills the current
process when the clock changes. Supervisor will then restart agent,
and we can recover.

It's not nice at all. Probably the result of an OS bug. Hopefully,
there will be a better solution in future.

"""

import os
import signal
from time import sleep, time
from threading import Thread


CHECK_PERIOD = 1.0
MAX_CLOCK_DISCREPENCY = 10.0


class ClockCheckThread(Thread):
    """Thread to restart agent when the clock changes."""

    def __init__(self):
        super(ClockCheckThread, self).__init__()
        self.daemon = True
        self.running = True

    def run(self):
        """Calls `on_fail` when the system clock changes."""
        while self.running:
            start = time()
            sleep(CHECK_PERIOD)
            elapsed = time() - start

            if elapsed > MAX_CLOCK_DISCREPENCY:
                self.on_fail()

    def on_fail(self):
        """Clock has changed, kill the current process."""
        # Supervisor will restart the agent.
        pid = os.getpid()
        os.kill(pid, signal.SIGKILL)



from __future__ import unicode_literals
from __future__ import print_function

from .compat import with_metaclass


registry = {}


class SubCommandMeta(type):
    """Keeps a registry of sub-commands"""

    def __new__(cls, name, base, attrs):
        new_class = type.__new__(cls, name, base, attrs)
        if name not in ("SubCommand", "SubCommandType"):
            registry[name.lower()] = new_class
        return new_class


class SubCommandType(object):
    """Base class for sub-commands"""

    __metaclass__ = SubCommandMeta

    def __init__(self, app):
        self.app = app

    def add_arguments(self, parser):
        pass

    def run(self):
        raise NotImplementedError


class SubCommand(with_metaclass(SubCommandMeta, SubCommandType)):
    pass



from __future__ import print_function
from __future__ import unicode_literals

import json
import logging
import sys
import threading
import time
from collections import defaultdict, deque

from lomond import WebSocket
from lomond.constants import USER_AGENT as LOMOND_USER_AGENT
from lomond.persist import persist
from lomond.errors import WebSocketError

from . import bencode
from . import packets
from ..compat import text_type
from .dispatcher import Dispatcher, expose
from .packets import M2MPacket as Packet
from .packets import PacketType
from .._version import __version__


log = logging.getLogger("m2m")


class ClientError(Exception):
    pass


class ChannelFile(object):
    """Mimicks a writable file."""

    def __init__(self, client, channel_no):
        self.client = client
        self.channel_no = channel_no

    def write(self, data):
        # http://stackoverflow.com/questions/23932332/writing-bytes-to-standard-output-in-a-way-compatible-with-both-python2-and-pyth
        # retrieve stdout as a binary file object
        output = getattr(sys.stdout, "buffer", sys.stdout)
        output.write(data)
        self.client.channel_write(self.channel_no, data)

    def fileno(self):
        return None


class Channel(object):
    """An interface to a channel."""

    def __init__(self, client, number):
        self.client = client
        self.number = number
        self._closed = False

        self._data_callback = None
        self._close_callback = None
        self._control_callback = None
        self._lock = threading.RLock()
        self.deque = deque()
        self._data_event = threading.Event()

    def __repr__(self):
        """Show the channel number."""
        return "<channel {}>".format(self.number)

    def close(self):
        """Call to *request* a close."""
        if not self._closed:
            self.client.close_channel(self.number)

    def on_close(self):
        """Called when the notify_close packet is received."""
        if self._closed:
            return
        self._closed = True
        try:
            if self._close_callback is not None:
                self._close_callback()
        except:
            log.exception("error in close callback")

    @property
    def is_closed(self):
        return self._closed

    def on_data(self, data):
        """On incoming data."""
        if self._closed:
            log.debug("%s bytes from closed %r ignored", len(data), self)
            return
        if self._data_callback is not None:
            self._data_callback(data)
        else:
            with self._lock:
                self.deque.append(data)
                self._data_event.set()

    def on_control(self, data):
        """On control data."""
        if self._closed:
            log.debug("%s bytes from closed %r ignored", len(data), self)
            return
        if self._control_callback is not None:
            self._control_callback(data)

    def set_callbacks(self, on_data=None, on_close=None, on_control=None):
        self._data_callback = on_data
        self._close_callback = on_close
        self._control_callback = on_control

    @property
    def size(self):
        with self._lock:
            return sum(len(b) for b in self.deque)

    def __nonzero__(self):
        return self._data_event.is_set()

    def __bool__(self):
        return self.__nonzero__()

    def read(self, count, timeout=None, block=False):
        """Read up to `count` bytes."""
        incoming_bytes = []
        bytes_remaining = count

        # Block until data
        if block:
            if not self._data_event.wait(timeout):
                return b""

        with self._lock:
            # Data may be spread across multiple / partial messages
            while self.deque and bytes_remaining:
                head = self.deque[0]
                read_bytes = min(bytes_remaining, len(head))
                incoming_bytes.append(head[:read_bytes])
                bytes_left = head[read_bytes:]
                bytes_remaining -= read_bytes
                if not bytes_left:
                    self.deque.popleft()
                else:
                    self.deque[0] = bytes_left
            if not self.deque:
                self._data_event.clear()

        return b"".join(incoming_bytes)

    def write(self, data):
        assert isinstance(data, bytes), "data must be bytes"
        if not self.is_closed:
            with self._lock:
                self.client.channel_write(self.number, data)

    def send_control(self, control):
        """Write a control packet."""
        if not self.is_closed:
            with self._lock:
                self.client.channel_control_write(self.number, control)

    def get_file(self):
        return ChannelFile(self.client, self.number)


class WSClient(threading.Thread):
    """Interface to the M2M server."""

    def __init__(
        self,
        manager,
        url,
        remote_directory,
        uuid=None,
        channel_callback=None,
        control_callback=None,
        **kwargs
    ):
        super(WSClient, self).__init__()
        self.manager = manager
        self.url = url
        self.remote_directory = remote_directory
        _user_agent = "Agent/{} {}".format(__version__, LOMOND_USER_AGENT)
        self.websocket = WebSocket(url, agent=_user_agent, compress=True)
        self.channel_callback = channel_callback
        self.control_callback = control_callback

        self._closed = False
        self.identity = uuid
        self.channels = {}
        self.last_packet_time = time.time()

        self.callback_lock = threading.RLock()
        self.write_lock = threading.Lock()
        self.callbacks = defaultdict(list)
        self.hooks = defaultdict(list)

        self.dispatcher = Dispatcher(packet_cls=Packet, handler_instance=self, log=log)

        self.name = "m2m"  # Thread name
        self.daemon = True

    def __repr__(self):
        """Return the URL."""
        return "WSClient({!r})".format(self.url)

    @property
    def is_closed(self):
        return self._closed

    @property
    def open_channels(self):
        """List of open channels."""
        return self.channels.keys()

    def add_callback(self, command_id, callback):
        self.callbacks[command_id].append(callback)

    def callback(self, command_id, result):
        with self.callback_lock:
            if command_id in self.callbacks:
                for callback in self.callbacks[command_id]:
                    try:
                        callback(result)
                    except Exception:
                        log.exception("error in command callback")
                del self.callbacks[command_id]

    def clear_callbacks(self):
        """Clear all callbacks, because they may be blocking."""
        with self.callback_lock:
            for command_id, callbacks in self.callbacks.items():
                for callback in callbacks:
                    try:
                        callback(None)
                    except Exception:
                        log.exception("error clearing callback")

    def get_channel(self, channel_no):
        # TODO: Create channels in response to packets
        if channel_no not in self.channels:
            self.channels[channel_no] = Channel(self, channel_no)
        return self.channels[channel_no]

    def has_channel(self, channel_no):
        return channel_no in self.channels

    def close_channel(self, channel_no):
        log.debug("request close")
        self.send("request_close", port=channel_no)

    def hard_close_channels(self):
        """Called when all the channels have been abruptly closed."""
        for channel in self.channels.values():
            channel.on_close()

    def run(self):
        """Main websocket handling loop."""
        try:
            with self.websocket:
                for event in persist(self.websocket):
                    log.debug("WS %r", event)
                    try:
                        self.on_event(event)
                    except Exception as error:
                        log.exception("error handling websocket event")
        except (SystemExit, KeyboardInterrupt):
            log.info("exit requested")
        except Exception:
            log.exception("unhandled error from websocket")
        self.on_close()

    def on_event(self, event):
        """Called when new websocket events arrive."""
        if event.name == "ready":
            self.on_ready()
        elif event.name == "disconnected":
            self.on_disconnected()
        elif event.name == "binary":
            self.on_binary(event.data)
        elif event.name == "poll":
            self.sync_identity()

    def close(self, timeout=5):
        self.websocket.close()
        self._closed = True
        self.identity = None

    def send(self, packet, *args, **kwargs):
        """Send a packet. Will encode if necessary."""
        if isinstance(packet, (bytes, text_type)):
            packet = PacketType[packet].value
        if isinstance(packet, (PacketType, int)):
            packet = Packet.create(packet, *args, **kwargs)
        if not getattr(packet, "no_log", False):
            log.debug("sending %r", packet)

        packet_bytes = packet.encode_binary()
        self.send_bytes(packet_bytes)

    def send_bytes(self, packet_bytes):
        """Send bytes over the websocket."""
        try:
            self.websocket.send_binary(packet_bytes)
        except WebSocketError:
            return False
        else:
            return True

    def on_ready(self):
        """Called when WS is opened."""
        log.debug("websocket opened")
        if self.identity is None:
            self.send(PacketType.request_join)
        else:
            self.send(PacketType.request_identify, uuid=self.identity)

    def on_binary(self, data):
        """On a WS message."""
        self.last_packet_time = time.time()
        try:
            packet = bencode.decode(data)
        except:
            log.exception("packet could not be decoded")
        else:
            self.on_packet(packet)

    def sync_identity(self):
        """Ask manager to set the m2m identity."""
        self.manager.set_identity(self.identity)

    def on_disconnected(self):
        """Called when ws socket closes."""
        self.identity = None
        self.clear_callbacks()
        self.hard_close_channels()

    def on_packet(self, packet):
        """Called with a binary packet."""
        try:
            packet_type = packet[0]
            packet_body = packet[1:]
        except Exception:
            log.exception("packet is badly formatted")
        else:
            try:
                packet_type = packets.PacketType(packet_type)
            except Exception:
                log.debug("unknown packet; %r", packet_type)
            else:
                self.dispatcher.dispatch(packet_type, packet_body)

    def channel_write(self, channel, data):
        """Write data to a virtual channel."""
        self.send(PacketType.request_send, channel=channel, data=data)

    def on_instruction(self, sender, data):
        """Called with an instruction."""
        log.debug("instruction from {%s} %r", sender, data)
        self.manager.on_instruction(sender, data)

    def channel_control_write(self, channel, control_dict):
        """Send a channel control packet."""
        control_json = json.dumps(control_dict)
        self.send(PacketType.request_send_control, channel=channel, data=control_json)

    # --------------------------------------------------------
    # Packet handlers
    # -------------------------------------------------------

    @expose(PacketType.null)
    def handle_null(self, packet_type):
        """Ignore null packet."""
        # Null packets may be sent just to check the connection

    @expose(PacketType.set_identity)
    def handle_set_identity(self, packet_type, identity):
        """Server is telling us about our identity."""
        if not self.is_closed:
            log.debug("setting identity to %s", identity)
            self.identity = identity
            self.sync_identity()

    @expose(PacketType.ping)
    def handle_ping(self, packet_type, data):
        """Ping send from the server, send back a pong with the same data."""
        self.send("pong", data=data[:1024])

    @expose(PacketType.welcome)
    def handle_welcome(self, packet_type):
        """Welcome packet means we can start talking to the m2m server."""
        pass

    @expose(PacketType.log)
    def handle_log(self, packet_type, msg):
        """The server has sent a message for us to write to the logs."""
        log.debug(msg)

    @expose(PacketType.route)
    def handle_route(self, packet_type, channel, data):
        """Route packet containing data to write to a 'channel'."""
        channel = self.get_channel(channel)
        if self.channel_callback is not None:
            try:
                self.channel_callback(channel, data)
            except Exception:
                log.exception("error in channel callback")
        channel.on_data(data)

    @expose(PacketType.route_control)
    def handle_route_control(self, packet_type, channel, data):
        """A control packet is out of band data associated with an existing channel."""
        channel = self.get_channel(channel)
        if self.control_callback is not None:
            try:
                self.control_callback(channel, data)
            except Exception:
                log.exception("error in channel callback")
        channel.on_control(data)

    @expose(PacketType.notify_open)
    def on_notify_open(self, packet_type, channel_no):
        """The server has told us of a new channel."""
        channel = self.get_channel(channel_no)
        log.debug("%s opened", channel)

    @expose(PacketType.notify_close)
    def on_notify_close(self, packet_type, channel_no):
        """The server has told us of a channel being closed."""
        log.debug("%s closed", channel_no)
        if self.has_channel(channel_no):
            channel = self.get_channel(channel_no)
            channel.on_close()
            del self.channels[channel_no]

    @expose(PacketType.notify_login_success)
    def on_login_success(self, packet_type, user):
        """Logged in users have special privileges (typically not needed by dpcore clients)."""
        self.user = user
        log.debug("logged in as %s", user)

    @expose(PacketType.response)
    def on_response(self, packet_type, command_id, result):
        """We have received a response to a command."""
        self.callback(command_id, result)

    @expose(PacketType.instruction)
    def on_instruction_packet(self, packet_type, sender, data):
        """An instruction packet contains application specific data."""
        try:
            self.on_instruction(sender, data)
        except Exception:
            log.exception("error handling instruction")

    @expose(PacketType.open_remote_file)
    def on_open_remote_file(self, packet_type, upload_id, path):
        # type: (PacketType, bytes, bytes) -> None
        """Server wants to open a remote file."""
        self.remote_directory.on_open_remote_file(self, upload_id, path)

    @expose(PacketType.close_remote_file)
    def on_close_remote_file(self, packet_type, upload_id):
        # type: (PacketType, bytes) -> None
        """Server wants to close a remote file."""
        self.remote_directory.on_close_remote_file(self, upload_id)

    @expose(PacketType.read_remote_file)
    def on_read_remote_file(self, packet_type, upload_id, offset, size):
        # type: (PacketType, bytes, int, int) -> None
        """Server wants to read from remote file."""
        self.remote_directory.on_read_remote_file(self, upload_id, offset, size)

    @expose(PacketType.scan_remote_directory)
    def on_scan_remote_directory(self, packet_type):
        # type: (PacketType) -> None
        """Server requests a scan of the remote directory."""

        def on_success():
            self.send("scan_remote_directory_result", "")

        try:
            self.remote_directory.scan(on_success=on_success)
        except Exception as error:
            self.send("scan_remote_directory_result", str(error))




"""
Python 2/3 compatibility layer

Base on the following post:

http://lucumr.pocoo.org/2013/5/21/porting-to-python-3-redux/

"""

import sys

# Use these flags for code for a particular PY version
PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

# Basic types and builtin
if PY2:
    text_type = unicode
    binary_type = str
    string_types = (basestring,)
    xrange = xrange
    unichr = unichr
    int_types = (int, long)
    py2bytes = lambda s: s.encode("utf-8")
    number_types = (int, long, float)
    next_method_name = "next"
    raw_input = raw_input
else:
    text_type = str
    binary_type = bytes
    xrange = range
    string_types = (str, bytes)
    unichr = chr
    int_types = (int,)
    py2bytes = lambda s: s
    number_types = (int, float)
    next_method_name = "__next__"
    raw_input = input

# Use these functions for iterating over keys / values / items
if PY2:
    iterkeys = lambda d: d.iterkeys()
    itervalues = lambda d: d.itervalues()
    iteritems = lambda d: d.iteritems()
else:
    iterkeys = lambda d: iter(d.keys())
    itervalues = lambda d: iter(d.values())
    iteritems = lambda d: iter(d.items())


# These functions have been re-arranges in PY3
if PY2:
    import urlparse
    from urlparse import urlparse, parse_qs, urlunparse
    from urllib import urlencode, quote
    from itertools import izip_longest as zip_longest
    from urllib2 import urlopen, HTTPError
    import Queue as queue
    from urllib2 import Request
else:
    from urllib.parse import urlparse, parse_qs, urlunparse
    from urllib.parse import urlencode, quote
    from itertools import zip_longest
    from urllib.request import urlopen, HTTPError
    import queue
    from urllib.request import Request


# pickle is the C version on PY3
if PY2:
    import cPickle as pickle
else:
    import pickle


# For classes that convert to a unicode string, return unicode from __str__
# and decorate with this function
if PY2:

    def implements_to_string(cls):
        cls.__unicode__ = cls.__str__
        cls.__str__ = lambda x: x.__unicode__().encode("utf-8")
        return cls


else:
    implements_to_string = lambda x: x

# To implement an iterator, call the next method __next__ and decorate the class
# with the following
if PY2:

    def implements_iterator(cls):
        cls.next = cls.__next__
        del cls.__next__
        return cls


else:
    implements_iterator = lambda x: x

# If a class converts to a bool, call the method __bool__, and decorate with this function
if PY2:

    def implements_bool(cls):
        cls.__nonzero__ = cls.__bool__
        del cls.__bool__
        return cls


else:

    def implements_bool(cls):
        return cls


def with_metaclass(meta, *bases):
    class metaclass(meta):
        __call__ = type.__call__
        __init__ = type.__init__

        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)

    return metaclass("temporary_class", None, {})



from __future__ import print_function, unicode_literals

import sys

import io

"""
Encode / Decode Bencode (http://en.wikipedia.org/wiki/Bencode)

"""


PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY3:
    number_types = (int,)
    text_type = str
else:
    number_types = (long, int)
    text_type = unicode


class EncodingError(ValueError):
    pass


class DecodeError(Exception):
    pass


class DecoderError(Exception):
    """Exception occurred with the data being decoded"""

    (
        PRECEDING_ZERO_IN_SIZE,
        MAX_SIZE_REACHED,
        ILLEGAL_DIGIT_IN_SIZE,
        ILLEGAL_DIGIT,
    ) = range(4)

    error_text = {
        PRECEDING_ZERO_IN_SIZE: "PRECEDING_ZERO_IN_SIZE",
        MAX_SIZE_REACHED: "MAX_SIZE_REACHED",
        ILLEGAL_DIGIT_IN_SIZE: "ILLEGAL_DIGIT_IN_SIZE",
        ILLEGAL_DIGIT: "ILLEGAL_DIGIT",
    }

    def __init__(self, code, text):
        self.code = code
        self.text = text
        super(DecoderError, self).__init__()

    def __str__(self):
        return "{} (#{}), {}".format(
            DecoderError.error_text[self.code], self.code, self.text
        )


def encode(obj):
    """Encode to Bencode, return bytes"""
    binary = []
    append = binary.append

    def add_encode(obj):
        if isinstance(obj, bytes):
            append("{}:".format(len(obj)).encode())
            append(obj)
        elif isinstance(obj, text_type):
            obj_bytes = obj.encode("utf-8")
            append("{}:".format(len(obj_bytes)).encode())
            append(obj_bytes)
        elif isinstance(obj, number_types):
            append("i{}e".format(obj).encode())
        elif isinstance(obj, (list, tuple)):
            append(b"l")
            for item in obj:
                add_encode(item)
            append(b"e")
        elif isinstance(obj, dict):
            append(b"d")
            keys = sorted(obj.keys())
            for k in keys:
                if not isinstance(k, bytes):
                    raise EncodingError("dict keys must be bytes")
                add_encode(k)
                add_encode(obj[k])
            append(b"e")
        else:
            raise EncodingError("value {!r} can not be encoded in Bencode".format(obj))

    add_encode(obj)
    bencode_data = b"".join(binary)
    assert decode(bencode_data)
    return bencode_data


def decode(data):
    """Decode Bencode, return an object."""
    assert isinstance(data, bytes), "decode takes bytes"
    return _decode(io.BytesIO(data).read)


def _decode(read):
    """Decode bencode, `read` should be a callable that returns number of bytes."""
    # TODO: Some input validation
    obj_type = read(1)
    if obj_type == b"":
        raise DecodeError("invalid input")
    if obj_type == b"e":
        return None
    if obj_type == b"i":
        number_bytes = b""
        while 1:
            c = read(1)
            if not c.isdigit() and c != b"-":
                if c != b"e":
                    raise DecodeError("illegal digit in size")
                break
            number_bytes += c
        number = int(number_bytes)
        return number
    elif obj_type == b"l":
        l = []
        while 1:
            i = _decode(read)
            if i is None:
                break
            l.append(i)
        return l
    elif obj_type == b"d":
        kv = []
        while 1:
            k = _decode(read)
            if k is None:
                break
            v = _decode(read)
            kv.append((k, v))
        return dict(kv)
    else:
        size_bytes = obj_type
        while 1:
            c = read(1)
            if c == b":":
                break
            size_bytes += c
        size = int(size_bytes)
        return read(size)



from __future__ import unicode_literals
from __future__ import print_function

"""
Packet management

"""

from . import bencode
from ..compat import int_types, binary_type, text_type, with_metaclass


class PacketError(Exception):
    """A packet format error."""


class PacketFormatError(PacketError):
    """Packet is badly formatted."""


class UnknownPacketError(PacketError):
    """A packet we don't know how to handle."""


class PacketMeta(type):
    """Maintains a registry of packet classes."""

    def __new__(mcs, name, bases, attrs):
        packet_cls = super(PacketMeta, mcs).__new__(mcs, name, bases, attrs)
        if bases[0] is not object:
            if packet_cls.type >= 0:
                assert (
                    packet_cls.type not in packet_cls.registry
                ), "packet type {!r} has already been registered".format(
                    packet_cls, type
                )
                packet_cls.registry[packet_cls.type] = packet_cls
        return packet_cls


class PacketBaseType(object):
    """Metaclass to register packet type."""

    registry = {}

    # Packet type
    type = -1  # Indicates it is a base packet class

    # Named attributes, if using default init_data
    attributes = []

    def __init__(self, *args, **kwargs):
        self.init_params(args, kwargs)
        self.validate()

    def __repr__(self):
        data = {}
        for attrib_name, _attrib_type in self.attributes:
            try:
                data[attrib_name] = getattr(self, attrib_name)
            except AttributeError:
                continue

        def summarize(key, value):
            if key == "password":
                return "********"
            if isinstance(value, binary_type) and len(value) > 32:
                remaining = len(value) - 32
                return "{!r} + {} bytes".format(value[:32], remaining)
            return repr(value)

        params = ", ".join(
            "{}={}".format(k, summarize(k, v)) for k, v in sorted(data.items())
        )
        return "{}({})".format(self.__class__.__name__, params)

    def process_packet_type(self, packet_type):
        return packet_type

    @classmethod
    def create(cls, packet_type, *args, **kwargs):
        """Dynamically create a packet from its type and parameters."""
        packet_cls = cls.registry.get(cls.process_packet_type(packet_type))
        if packet_cls is None:
            raise ValueError("no packet type {}".format(packet_type))
        return packet_cls(*args, **kwargs)

    @classmethod
    def from_bytes(cls, packet_bytes):
        """Return a packet from a bytes string."""
        try:
            packet_data = bencode.decode(packet_bytes)
        except bencode.DecodeError as e:
            raise PacketFormatError("packet is badly formated ({!r})".format(e))

        if not isinstance(packet_data, list):
            raise PacketFormatError("packet must be a list")
        packet_type = packet_data[0]
        if not isinstance(packet_type, int_types):
            raise PacketFormatError("first value must be an integer")
        packet_body = packet_data[1:]
        try:
            packet_cls = cls.registry[packet_type]
        except:
            raise UnknownPacketError("unknown packet type '{}'".format(packet_type))
        return packet_cls.from_body(packet_body)

    @classmethod
    def from_body(cls, packet_body):
        """Return a packet object from a packet body."""
        params = {}
        for (attrib_name, attrib_type), value in zip(cls.attributes, packet_body):
            params[attrib_name] = value
        return cls(**params)

    @property
    def kwargs(self):
        return {
            attrib_name: getattr(self, attrib_name)
            for attrib_name, attrib_type in self.attributes
        }

    def get_method_args(self, arg_count):
        args = []
        kwargs = self.kwargs.copy()
        for attrib_name, _ in self.attributes[:arg_count]:
            args.append(kwargs.pop(attrib_name))
        return args, kwargs

    def init_params(self, args, kwargs):
        """Initialize from parameters."""
        # Default implementation copies named attributes

        params = {}

        for arg, (attribute_name, attrib_type) in zip(args, self.attributes):
            if isinstance(arg, text_type):
                arg = arg.encode("utf-8")
            params[attribute_name] = arg
        for k, v in kwargs.items():
            if isinstance(v, text_type):
                v = v.encode("utf-8")
            params[k] = v

        for attrib_name, attrib_type in self.attributes:
            if attrib_name not in params:
                raise PacketFormatError(
                    "missing attribute '{}', in {!r}".format(attrib_name, self)
                )
            value = params[attrib_name]
            if attrib_type is not None and not isinstance(value, attrib_type):
                raise PacketFormatError(
                    "parameter '{}' should be a {!r}, in {!r} (not {!r})".format(
                        attrib_name, attrib_type, self, type(value)
                    )
                )
            setattr(self, attrib_name, params[attrib_name])

    def validate(self):
        """Check packet data for errors."""
        pass

    @property
    def as_bytes(self):
        """Encode the packet as bytes."""
        return self.encode_binary()

    def encode_binary(self):
        """Encode the packet in to a byte string."""
        return bencode.encode(self.encode())

    def encode(self):
        """Encode the packet (including type header)."""
        data = [int(self.type)] + [
            getattr(self, attrib_name) for attrib_name, attrib_type in self.attributes
        ]
        return data


class PacketBase(with_metaclass(PacketMeta, PacketBaseType)):
    """Base class for packets."""



"""
A threadsafe object to enforce limits.

Similar in purpose to a Semaphore, but simpler because we never want to block. 

"""
from contextlib import contextmanager
import logging
from threading import RLock

log = logging.getLogger("m2m")


class LimiterError(Exception):
    """Base class for counter errors."""


class LimitReached(Exception):
    """Count has reached the maximum"""


@contextmanager
def limiter_context(limiter):
    """Context manager which increments the limiter, and decrements if there is an error."""
    # Increment the counter
    limiter.increment()
    with limiter._lock:
        try:
            # run setup code, while locked
            yield
        except Exception:
            # Decrement counter if any errors
            limiter.decrement()
            raise


class Limiter(object):
    """A thread safe counter with an upper limit."""

    def __init__(self, name, limit):
        """Create limiter object.

        Args:
            name (str): Name of limiter (used in error messages)
            limit (int): Upper limit.
        """
        assert limit > 0
        self._lock = RLock()
        self.name = name
        self._limit = limit
        self._value = 0

    def __repr__(self):
        return "<limiter {!r} {}/{}>".format(self.name, self._value, self._limit)

    def __call__(self):
        """Countext manager to increment limiter and decrement on error."""
        return limiter_context(self)

    def increment(self):
        """Increment the count.

        Raises a LimitReached exception if the increment would go above the limit.
        """
        with self._lock:
            if self._value >= self._limit:
                log.warning("%r reached limit", self)
                raise LimitReached(
                    "{} limit ({}) reached".format(self.name, self._limit)
                )
            self._value += 1
            log.debug("%r incremented", self)

    def decrement(self):
        """Decrement the count."""
        with self._lock:
            if self._value <= 0:
                # If this occurs it indicates a bug in the caller
                raise LimiterError("Counter can't be decremented below 0")
            self._value -= 1
            log.debug("%r decremented", self)



from __future__ import unicode_literals
from __future__ import print_function

"""
Dispatches incoming packets

"""


from ..compat import text_type

import logging
import inspect


class PacketFormatError(Exception):
    pass


class UnknownMethodError(Exception):
    """Unknown packet type"""


def expose(packet_type):
    def deco(f):
        f._dispatcher_exposed = True
        f._dispatcher_packet_type = packet_type
        return f

    return deco


class Dispatcher(object):
    """
    Base class to dispatch to handlers for a packet.

    May also be used to dispatch to methods of another object, rather than a base class.

    """

    def __init__(self, packet_cls=None, handler_instance=None, log=None):
        super(Dispatcher, self).__init__()
        self._handler_instance = handler_instance or self

        if log is None:
            self.log = logging.getLogger("dispatcher")
        else:
            self.log = log

        self._packet_cls = packet_cls
        self._packet_handlers = {}
        self._init_dispatcher()
        self._dispatch_enabled = True

    def set_packet_class(self, packet_cls):
        self._packet_cls = packet_cls

    def disable(self):
        """Prevent all further packet dispatch."""
        self._dispatch_enabled = False
        self._packet_handlers.clear()

    def _init_dispatcher(self):
        for method_name in dir(self._handler_instance):
            if method_name.startswith("_"):
                continue
            method = getattr(self._handler_instance, method_name)
            if getattr(method, "_dispatcher_exposed", False):
                packet_type = method._dispatcher_packet_type
                self._packet_handlers[packet_type] = method

    def dispatch(self, packet_type, packet_body):
        """Dispatch a packet to appropriate handler"""
        if not self._dispatch_enabled:
            return
        if not isinstance(packet_type, int):
            raise PacketFormatError("packet type should be an int")
        assert (
            self._packet_cls is not None
        ), "packet class must be set with set_packet_class"

        packet = self._packet_cls.create(packet_type, *packet_body)
        return self.dispatch_packet(packet)

    def dispatch_packet(self, packet):
        if not getattr(packet, "no_log", False):
            self.log.debug("received %r", packet)
        packet_type = packet.type
        method = self._packet_handlers.get(packet_type, None)

        if method is None:
            self.on_missing_handler(packet)
            return None

        arg_spec = inspect.getargspec(method)
        args, kwargs = packet.get_method_args(len(arg_spec[0]))

        try:
            inspect.getcallargs(method, packet_type, *args, **kwargs)
        except TypeError as e:
            raise PacketFormatError(text_type(e))

        try:
            ret = method(packet_type, *args, **kwargs)
        except Exception:
            raise
        else:
            return ret

    def on_missing_handler(self, packet):
        """Called when no handler is available to handle `packet`"""
        self.log.debug("missing handler for %r", packet)



from __future__ import unicode_literals
from __future__ import print_function

from ..subcommand import SubCommand
from .. import __version__

import sys


class Version(SubCommand):
    """Write version to stdout."""

    help = """Write version to stdout."""

    def run(self):
        sys.stdout.write(__version__)



import logging
import shutil
import os
import os.path
import tempfile
import typing

if typing.TYPE_CHECKING:
    from typing import Callable, Text

from . import disk_tools
from .directory_scanner import DirectoryScanner
from .m2m.packets import PacketType
from .m2m.wsclient import WSClient

# Amount of free space (bytes) to leave on disk
FREE_SPACE_MARGIN = 50 * 1000 * 1000


log = logging.getLogger("agent")


class RemoteDirectoryError(Exception):
    """Base exception for remote directory errors."""


class AddFail(RemoteDirectoryError):
    """Unable to add new upload."""


class ReadFail(RemoteDirectoryError):
    """Read operation failed (unlikely be recoverable)"""


class LowDiskSpace(RemoteDirectoryError):
    """There is not enough disk space to create a snapshot."""


class IllegalPath(RemoteDirectoryError):
    """The upload path is invalid."""


def validate_path(path):
    # type: (Text) -> bool
    """Check a path has no backrefs that may end up serving outside the remote directory root.

    Args:
        path (Text): A path.

    Returns:
        bool: True if the path is valid, otherwise False
    """
    return ".." not in path.split(os.sep)


class RemoteDirectory(object):
    """Manages remote directory."""

    def __init__(self, path, directory_scanner):
        # type: (Text, DirectoryScanner) -> None
        path = os.path.expanduser(path)
        self.path = os.path.abspath(path)
        self.directory_scanner = directory_scanner
        self.temp_path = tempfile.gettempdir()

    def __repr__(self):
        # type: () -> Text
        return "RemoteDirectory(%r)" % self.path

    def scan(self, on_success=None):
        # type: (Callable[[], None]) -> None
        """Scan remote directory."""
        self.directory_scanner.perform_scan(on_success=on_success)

    def get_snapshot_path(self, upload_id):
        # type: (Text) -> Text
        """Get the path to a snapshot.

        Args:
            upload_id (str): Upload ID.

        Returns:
            str: A path to the snapshot file.
        """
        try:
            temp_path = tempfile.gettempdir()
            dir_path = os.path.join(temp_path, "dataplicity_remote")
            if not os.path.isdir(dir_path):
                os.mkdir(dir_path)
                log.debug("%r created %s", self, dir_path)
            snapshot_path = os.path.join(dir_path, upload_id)
            return snapshot_path
        except Exception as error:
            log.warning("unable to get snapshot path; %r", error)
            raise RemoteDirectoryError("Failed to get snapshot path")

    def add_upload(self, path, upload_id):
        # type: (Text, Text) -> int
        """Add a new upload.

        Args:
            path (str): Path to the file to upload relative to self.path
            upload_id (str): Upload ID.

        Raises:
            LowDiskSpace: If there is not enough disk space to create a snapshot (including margin)
            AddFail: If the upload could not be added for any reason.

        Returns:
            int: The size of the snapshot file (in bytes)
        """

        if path.startswith("\0"):
            if path == "\0scan.json":
                # Special file served from temp
                file_path = os.path.join(
                    tempfile.gettempdir(), "__dataplicity_remote_directory_scan___.json"
                )
            else:
                raise IllegalPath("Unknown special path; %r" % path)
        else:
            # Regular file in remote directory
            file_path = os.path.join(self.path, path.lstrip("/"))

            if not validate_path(file_path) or not file_path.startswith(self.path):
                raise IllegalPath("Path %s is illegal" % file_path)

        try:
            disk_usage = disk_tools.disk_usage(self.temp_path)
            file_size = os.path.getsize(file_path)
        except Exception as error:
            # Don't want to fail here, probably worth honoring the request
            log.warning("failed to get disk usage information; %s", error)
        else:
            if disk_usage.used + file_size > disk_usage.total - FREE_SPACE_MARGIN:
                raise LowDiskSpace(
                    "not enough space on %s to create snapshot" % self.temp_path
                )

        snapshot_path = self.get_snapshot_path(upload_id)
        # Sanity check for path
        if not snapshot_path.startswith(self.temp_path):
            raise IllegalPath("Path %s is illegal" % path)

        try:
            # Copy file to temporary location, creating a "snapshot"
            shutil.copyfile(file_path, snapshot_path)
            snapshot_size = os.path.getsize(snapshot_path)
            log.debug("created snapshot for %s; %s", file_path, snapshot_path)
            return snapshot_size
        except Exception as error:
            log.error("failed to add_upload; %r", error)
            raise AddFail("Unable to add upload; %s" % error)

    def read_upload(self, upload_id, offset, size):
        # type: (Text, int, int) -> bytes
        """Read data from the snapshot.

        Args:
            upload_id (str): Upload ID.
            offset (int): Offset in the file.
            size (int): Number of bytes to read.

        Raises:
            ReadFail: If the file could not be read from.

        Returns:
            bytes: Raw data from the file.
        """
        snapshot_path = self.get_snapshot_path(upload_id)
        try:
            with open(snapshot_path, "rb") as snapshot_file:
                snapshot_file.seek(offset)
                data = snapshot_file.read(size)
                log.debug("read %i bytes from %s", len(data), snapshot_path)
        except Exception as error:
            # Most likely, the snapshot file has been removed
            log.error("failed to read_upload; %r", error)
            raise ReadFail("unable to read %r" % upload_id)
        return data

    def close_upload(self, upload_id):
        # type: (Text) -> None
        """Close the upload and remove the snapshot file.

        Args:
            upload_id (str): Upload ID.
        """
        snapshot_path = self.get_snapshot_path(upload_id)
        # If its already been removed, this is a no-opp
        if os.path.exists(snapshot_path):
            try:
                os.remove(snapshot_path)
            except Exception as error:
                log.warning("failed to remove %r; %r", snapshot_path, error)

    def on_open_remote_file(self, client, upload_id, path):
        # type: (WSClient, bytes, bytes) -> None
        """Open remote file packet."""
        try:
            size = self.add_upload(path.decode("utf-8"), upload_id.decode("utf-8"))
        except RemoteDirectoryError as error:
            # Known fail that we can report to the user
            client.send(
                PacketType.open_remote_file_result,
                upload_id=upload_id,
                size=-1,
                fail=1,
                fail_reason=str(error),
            )
        except Exception as error:
            client.send(
                PacketType.open_remote_file_result,
                upload_id=upload_id,
                size=-1,
                fail=1,
                fail_reason="open failed, see dataplicity.log",
            )
            log.exception("failed to add upload")
        else:
            client.send(PacketType.open_remote_file_result, upload_id, size, 0, "")

    def on_close_remote_file(self, client, upload_id):
        # type: (WSClient, bytes) -> None
        """On close remote file packet."""
        try:
            self.close_upload(upload_id.decode("utf-8"))
        except Exception:
            log.exception("failed to close upload")

    def on_read_remote_file(self, client, upload_id, offset, size):
        # type: (WSClient, bytes, int, int) -> None
        try:
            data = self.read_upload(upload_id.decode("utf-8"), offset, size)
        except RemoteDirectoryError as error:
            client.send(
                PacketType.read_remote_file_result,
                upload_id=upload_id,
                offset=offset,
                size=size,
                data="",
                fail=1,
                fail_reason=str(error),
            )
        except Exception:
            log.exception("failed to read upload")
            client.send(
                PacketType.read_remote_file_result,
                upload_id=upload_id,
                offset=offset,
                size=size,
                data="",
                fail=1,
                fail_reason="read failed, see dataplicity.log",
            )
        else:
            client.send(
                PacketType.read_remote_file_result,
                upload_id=upload_id,
                offset=offset,
                size=size,
                data=data,
                fail=0,
                fail_reason="",
            )



from __future__ import unicode_literals

import os
import subprocess
import re
import logging

log = logging.getLogger("agent")


class TagError(Exception):
    """Custom exception raised when get_tag_list has an exception"""


def get_tag_list():
    """Run the dataplicity.tags script, get output as a list of tags"""

    home_dir = os.environ.get("HOME", "/home/dataplicity/")
    tag_executable = os.path.join(home_dir, "dataplicity_tags")

    # Early out if the script isn't there.
    if not os.path.exists(tag_executable):
        log.debug("tag executable %s does not exist", tag_executable)
        return []

    log.debug("reading tags from %s", tag_executable)
    try:
        output = subprocess.check_output(tag_executable)
    except OSError as error:
        log.debug("failed to run %s; %s", tag_executable, error)
        return []
    except Exception as error:
        log.error("error running %s; %s", tag_executable, error)
        raise TagError("error running %s" % tag_executable)

    str_output = output.decode("utf-8", errors="ignore")

    # regex split on comma, spaces, newline and tabs
    tag_list = re.split(r"[,\s\n\t]", str_output)
    tags = [tag.strip()[:25] for tag in tag_list if tag]
    return tags



"""

An M2M Service to retrieve files.

File data is sent in chunks.

"""

from __future__ import print_function
from __future__ import unicode_literals

from functools import partial
import logging
import os.path
import threading

from lomond.errors import WebSocketError

from ..limiter import LimitReached
from ..constants import CHUNK_SIZE, SERVER_BUSY


log = logging.getLogger("m2m")


class FileService(threading.Thread):
    """A thread the sends data from a file over m2m."""

    # A word on lifetime management of this and similar objects...
    # There does not need to be any other references to these objects;
    # the thread maintains a reference to a m2m channel in the run
    # function, so the entire object will be garbage collected when
    # the thread exits.

    def __init__(self, limiter, channel, path):
        self.limiter = limiter
        self._repr = "FileService({!r}, {!r})".format(channel, path)
        try:
            with limiter():
                super(FileService, self).__init__(
                    args=(channel, path), target=self.run_service
                )
                self.start()
        except Exception as error:
            # Could be limit reached, or out of threads
            log.warning("failed to create file service; %s", error)
            channel.write(SERVER_BUSY)
            channel.close()

    def __repr__(self):
        return self._repr

    @classmethod
    def send_error(cls, channel, status, msg, **extra):
        """Send a control packet with an error"""
        error = {
            "service": "remote-file",
            "type": "error",
            "status": status,
            "msg": msg,
        }
        error.update(extra)
        channel.send_control(error)

    def run_service(self, channel, path):
        """Run the thread and log exceptions."""
        try:
            try:
                self._run_service(channel, path)
            except Exception:
                log.exception("error running %r", self)
                self.send_error(channel, "error", "internal error, see agent logs")
        finally:
            self.limiter.decrement()

    def _run_service(self, channel, path):
        """Send a file over a port."""
        log.debug("%r started", self)
        bytes_sent = 0
        if not path.startswith("/"):
            path = "/" + path
        try:
            with open(path, "rb") as read_file:
                read = partial(read_file.read, CHUNK_SIZE)
                for chunk in iter(read, b""):
                    if channel.is_closed:
                        log.warning("%r m2m closed prematurely", self)
                        break
                    channel.write(chunk)
                    bytes_sent += len(chunk)
        except IOError as error:
            self.send_error(channel, "ioerror", msg="unable to open file")
            log.debug('unable to read file "%s"; %r', path, error)
        except WebSocketError as websocket_error:
            log.warning("websocket error (%s)", websocket_error)
        except Exception:
            self.send_error(channel, "error", "internal error, see agent logs")
            log.exception("error in file service")
        else:
            log.info('read %s byte(s) from "%s"', bytes_sent, path)
        finally:
            channel.close()



"""Manage a subprocess that streams to a remote side"""

from __future__ import unicode_literals
from __future__ import print_function

import json
import logging
import os
import signal
import shlex
import time
from threading import Thread

from . import proxy

log = logging.getLogger("m2m")


class PidWaitThread(Thread):
    """A thread to wait for a process to exit, and display warnings."""

    def __init__(self, command, pid):
        self.command = command
        self.pid = pid
        super(PidWaitThread, self).__init__()
        self.daemon = True

    def __str__(self):
        return '"%s" (pid=%i)' % (self.command, self.pid)

    def run(self):
        """Poll PID for exit status, displaying warnings at given intervals if it is taking too long."""
        log.debug("Waiting for %s to close", self)
        start_time = time.time()
        warnings = [
            5,
            10,
            30,
            60,
            60 * 10,  # ten minutes
            60 * 60,  # An hour
            60 * 60 * 24,  # A day
        ]  # Seconds between warnings
        sent_kill = False
        kill_time = 15  # Send a sigkill if process doesn't shut down (in seconds)
        while warnings:
            time.sleep(2)
            # os.WNOHANG flags makes waitpid return immediately if the process is running
            pid, exit_code = os.waitpid(self.pid, os.WNOHANG)
            if pid:
                # The process has returned an exit code
                log.debug("process %s exited with code=%i", self, exit_code)
                break
            else:
                # Process is still running
                time_passed = time.time() - start_time
                if time_passed >= warnings[0]:
                    warnings.pop(0)
                    log.warning(
                        "process %s failed to exit after %.1f seconds",
                        self,
                        time_passed,
                    )
                if not sent_kill and time_passed >= kill_time:
                    sent_kill = True
                    log.debug("sending SIGKILL to process %s", self)
                    os.kill(self.pid, signal.SIGKILL)
        else:
            log.error("process %s will not die!", self)


class RemoteProcess(proxy.Interceptor):
    """Process managed remotely over m2m."""

    def __init__(self, limiter, command, channel, user=None, group=None, size=None):
        self.limiter = limiter
        self.command = command
        self.channel = channel
        self.size = size
        self._closed = False
        self.channel.set_callbacks(
            on_data=self.on_data, on_close=self.on_close, on_control=self.on_control
        )

        super(RemoteProcess, self).__init__(user=user, group=group, size=size)

    @property
    def is_closed(self):
        return self._closed

    def __repr__(self):
        return "RemoteProcess({!r}, {!r}, pid={})".format(
            self.command, self.channel, self.pid
        )

    def run(self):
        try:
            self.spawn(shlex.split(self.command))
        finally:
            self.limiter.decrement()

    def on_data(self, data):
        try:
            self.stdin_read(data)
        except Exception:
            self.channel.close()

    def on_control(self, data):
        try:
            control = json.loads(data)
        except Exception:
            log.exception("error decoding control")
            return
        control_type = control.get("type", None)
        if control_type == "window_resize":
            size = control["size"]
            log.debug("resize terminal to {} X {}".format(*size))
            self.resize_terminal(size)
        else:
            log.warning("unknown control packet {}".format(control_type))

    def on_close(self):
        self.close()

    def master_read(self, data):
        self.channel.write(data)
        super(RemoteProcess, self).master_read(data)

    def write_master(self, data):
        super(RemoteProcess, self).write_master(data)

    def close(self):
        if not self._closed and self.pid is not None:
            log.debug("sending SIGHUP to %r", self)
            os.kill(self.pid, signal.SIGHUP)
            try:
                wait_thread = PidWaitThread(self.command, self.pid)
                wait_thread.start()
            except Exception as error:
                log.warning("pid wait thread failed to launch; %s", error)
        self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()



"""Manage packet types / encoding."""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

from enum import IntEnum, unique
from .packetbase import PacketBase
from ..compat import text_type, int_types


@unique
class PacketType(IntEnum):
    """The top level packet type."""

    # Null packet, does nothing
    null = 0

    # Client sends this to join the server
    request_join = 1

    # Client sends this to re-connect
    request_identify = 2

    # Sent by the server if request_join or request_identity was successful
    welcome = 3

    # Textual information for developer
    log = 4

    # Send a packet to another node
    request_send = 5

    # Incoming data from the server
    route = 6

    # Send the packet back
    ping = 7

    # A ping return
    pong = 8

    # Set the clients identity
    set_identity = 9

    # Open a channel
    request_open = 10

    # Close a channel
    request_close = 11

    # Close all channels
    request_close_all = 12

    # Keep alive packet
    keep_alive = 13

    # Sent by the server to notify a client that a channel has been opened
    notify_open = 14

    # Request login for privileged accounts
    request_login = 15

    instruction = 16

    notify_login_success = 17

    notify_login_fail = 18

    # Notify the client that a channel has closed
    notify_close = 19

    # Client wishes to disconnect
    request_leave = 20

    # An out of band route control packet
    route_control = 21

    # Send a route_control packet
    request_send_control = 22

    # Notify client of name # ADDED 15/11/16
    # notify_name = 23

    # Remote directory files ADDED 25/2/21
    open_remote_file = 24
    open_remote_file_result = 25
    close_remote_file = 26
    read_remote_file = 27
    read_remote_file_result = 28
    scan_remote_directory = 29
    scan_remote_directory_result = 30

    response = 100

    command_add_route = 101

    command_send_instruction = 102

    command_log = 103

    command_broadcast_log = 104

    # command_forward = 105

    command_set_name = 106

    command_check_nodes = 107

    command_get_identities = 108

    command_set_auth = 109

    # Associate meta info with a node
    command_set_meta = 110

    # Get Meta info
    command_get_meta = 111

    peer_add_route = 200

    peer_forward = 201

    peer_notify_disconnect = 202

    # Tell peers about a named connection
    peer_notify_name = 203

    peer_close_port = 204


class M2MPacket(PacketBase):
    """Base class, not a real packet."""

    type = -1

    registry = {}

    @classmethod
    def process_packet_type(cls, packet_type):
        """Enable the use of strings to identify packets."""
        if isinstance(packet_type, (bytes, text_type)):
            packet_type = PacketType[packet_type].value
        return int(packet_type)


# ------------------------------------------------------------
# Packet classes
# ------------------------------------------------------------


class NullPacket(M2MPacket):
    """Probably never sent, this may be used as a sentinel at some point."""

    type = PacketType.null


class RequestJoinPacket(M2MPacket):
    """Client requests joining the server."""

    type = PacketType.request_join


class RequestIdentifyPacket(M2MPacket):
    """Client requests joining the server with a particular identity."""

    type = PacketType.request_identify
    attributes = [("uuid", bytes)]


class WelcomePacket(M2MPacket):
    """Send to the client when an identity has been recorded."""

    type = PacketType.welcome


class LogPacket(M2MPacket):
    """Log information, client may ignore."""

    type = PacketType.log
    attributes = [("text", bytes)]


class RequestSendPacket(M2MPacket):
    """Request to send data to a connection."""

    type = PacketType.request_send
    attributes = [("channel", int_types), ("data", bytes)]


class KeepAlivePacket(M2MPacket):
    """Keep alive packet."""

    type = PacketType.keep_alive


class RoutePacket(M2MPacket):
    """Route data."""

    type = PacketType.route
    attributes = [("channel", int_types), ("data", bytes)]


class RouteControlPacket(M2MPacket):
    """Route data."""

    type = PacketType.route_control
    attributes = [("channel", int_types), ("data", bytes)]


class RequestSendControlPacket(M2MPacket):
    """Request to send data to a connection."""

    type = PacketType.request_send_control
    attributes = [("channel", int_types), ("data", bytes)]


class PingPacket(M2MPacket):
    """Ping packet to check connection."""

    type = PacketType.ping
    attributes = [("data", bytes)]


class PongPacket(M2MPacket):
    """Response to Ping packet."""

    type = PacketType.pong
    attributes = [("data", bytes)]


class SetIdentityPacket(M2MPacket):
    """
    Sets a client's identity.

    This was for debugging, clients are sent an identity normally.

    """

    type = PacketType.set_identity
    attributes = [("uuid", bytes)]


class NotifyOpenPacket(M2MPacket):
    """Let the client know a channel was opened."""

    type = PacketType.notify_open
    attributes = [("channel", int_types)]


class RequestLoginPacket(M2MPacket):
    """Login for extra privileges."""

    type = PacketType.request_login
    attributes = [("username", bytes), ("password", bytes)]


class NotifyLoginSuccessPacket(M2MPacket):
    """Login success."""

    type = PacketType.notify_login_success
    attributes = [("user", bytes)]


class NotifyLoginFailPacket(M2MPacket):
    """Login failed."""

    type = PacketType.notify_login_fail
    attributes = [("message", bytes)]


class NotifyClosePacket(M2MPacket):
    """channel was closed."""

    type = PacketType.notify_close
    attributes = [("port", int)]


class RequestClosePacket(M2MPacket):
    """Ask server to close a port."""

    type = PacketType.request_close
    attributes = [("port", int)]


class RequestLeavePacket(M2MPacket):
    """Polite way of disconnecting from the server."""

    type = PacketType.request_leave


class InstructionPacket(M2MPacket):
    """Send an 'instruction' which is an application define packet not send through a channel."""

    type = PacketType.instruction
    attributes = [("sender", bytes), ("data", dict)]


class CommandResponsePacket(M2MPacket):
    """Sent in response to a command."""

    type = PacketType.response
    attributes = [("command_id", int_types), ("result", dict)]


class CommandAddRoutePacket(M2MPacket):
    """Command the server to generate a route from uuid1 to uuid2."""

    type = PacketType.command_add_route
    attributes = [
        ("command_id", int_types),
        ("node1", bytes),
        ("port1", int_types),
        ("node2", bytes),
        ("port2", int_types),
        ("requester", bytes),
        ("forwarded", int_types),
    ]


class CommandSendInstructionPacket(M2MPacket):
    """Send an instruction to a client."""

    type = PacketType.command_send_instruction
    attributes = [("command_id", int_types), ("node", bytes), ("data", dict)]


class CommandLogPacket(M2MPacket):
    """Send a message to be written to the logs."""

    type = PacketType.command_log
    attributes = [("command_id", int_types), ("node", bytes), ("text", bytes)]


class CommandBroadcastLogPacket(M2MPacket):
    """Send a message to all clients."""

    # Probably just for debug. Not sure what would happen with 1000s of clients
    type = PacketType.command_broadcast_log
    attributes = [("command_id", int_types), ("text", bytes)]


class CommandSetName(M2MPacket):
    """Set an alternative name of a node."""

    type = PacketType.command_set_name
    attributes = [("command_id", int_types), ("node", bytes), ("name", bytes)]


class CommandCheckNodes(M2MPacket):
    """Get identities from a list of names."""

    type = PacketType.command_check_nodes
    attributes = [("command_id", int_types), ("nodes", list)]


class CommandGetIdentities(M2MPacket):
    """Get identities from a list of names."""

    type = PacketType.command_get_identities
    attributes = [("command_id", int_types), ("nodes", list)]


class CommandSetAuth(M2MPacket):
    """Set auth information."""

    type = PacketType.command_set_auth
    attributes = [("command_id", int_types), ("expire", int_types), ("value", bytes)]


class CommandSetMeta(M2MPacket):
    """Set a meta key/value associated with a node."""

    type = PacketType.command_set_meta
    attributes = [
        ("command_id", int_types),
        ("requester", bytes),
        ("node", bytes),
        ("key", bytes),
        ("value", bytes),
    ]


class CommandGetMeta(M2MPacket):
    """Get a dictionary of meta values associated with a node."""

    type = PacketType.command_get_meta
    attributes = [("command_id", int_types), ("requester", bytes), ("node", bytes)]


class PeerAddRoutePacket(M2MPacket):
    """Tell the peer cluster about a route."""

    type = PacketType.peer_add_route
    attributes = [
        ("command_id", int),
        ("requester", bytes),
        ("node1", bytes),
        ("port1", int_types),
        ("node2", bytes),
        ("port2", int_types),
    ]


class PeerForwardPacket(M2MPacket):
    """Forward a packet to another node."""

    type = PacketType.peer_forward
    attributes = [("recipient", bytes), ("packet", bytes)]


class PeerNotifyDisconnect(M2MPacket):
    """Tell the peer cluster a client disconnected."""

    type = PacketType.peer_notify_disconnect
    attributes = [("node", bytes)]


class PeerNotifyName(M2MPacket):
    """Notify peer(s) about a named connection."""

    type = PacketType.peer_notify_name
    attributes = [("node", bytes), ("name", bytes)]


class PeerClosePort(M2MPacket):
    """Tell peer about a closed port."""

    type = PacketType.peer_close_port
    attributes = [("node", bytes), ("port", int)]


class OpenRemoteFile(M2MPacket):
    """Server requests opening a remote file."""

    type = PacketType.open_remote_file
    attributes = [("upload_id", bytes), ("path", bytes)]


class OpenRemoteFileResult(M2MPacket):
    """Tell server about the opening of a remote file."""

    type = PacketType.open_remote_file_result
    attributes = [
        ("upload_id", bytes),
        ("size", int_types),
        ("fail", int),
        ("fail_reason", bytes),
    ]


class CloseRemoteFile(M2MPacket):
    """Server requests close of remote file."""

    type = PacketType.close_remote_file
    attributes = [("upload_id", bytes)]


class ReadRemoteFile(M2MPacket):
    """Server requests read from remote file."""

    type = PacketType.read_remote_file
    attributes = [("upload_id", bytes), ("offset", int), ("size", int)]


class ReadRemoteFileResult(M2MPacket):
    """Tell server about result of read."""

    type = PacketType.read_remote_file_result
    attributes = [
        ("upload_id", bytes),
        ("offset", int),
        ("size", int),
        ("data", bytes),
        ("fail", int),
        ("fail_reason", bytes),
    ]


class ScanRemoteDirectory(M2MPacket):
    """Server wants us to rescan the remote directory."""

    type = PacketType.scan_remote_directory
    attributes = []


class ScanRemoteDirectoryResult(M2MPacket):
    """Inform server about result of directory scan."""

    type = PacketType.scan_remote_directory_result
    attributes = [("error", bytes)]



# Copyright (c) 2011 Joshua D. Bartlett
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import array
import fcntl
import logging
import os
import pty
import select
import sys
import termios
import pwd
import grp


# The following escape codes are xterm codes.
# See http://rtfm.etla.org/xterm/ctlseq.html for more.
START_ALTERNATE_MODE = set("\x1b[?{0}h".format(i) for i in ("1049", "47", "1047"))
END_ALTERNATE_MODE = set("\x1b[?{0}l".format(i) for i in ("1049", "47", "1047"))
ALTERNATE_MODE_FLAGS = tuple(START_ALTERNATE_MODE) + tuple(END_ALTERNATE_MODE)


log = logging.getLogger("dataplicity.m2m")


class Interceptor(object):
    """
    This class does the actual work of the pseudo terminal. The spawn() function is the main entrypoint.
    """

    def __init__(self, user=None, group=None, size=None):
        self.user = user
        self.group = group
        if size is None:
            size = [80, 24]
        self.size = size
        self.master_fd = None
        self.pid = None

    def spawn(self, argv=None):
        """
        Create a spawned process.
        Based on the code for pty.spawn().
        """
        assert self.master_fd is None

        pid, master_fd = pty.fork()
        self.master_fd = master_fd
        self.pid = pid
        if pid == pty.CHILD:
            if self.user is not None:
                try:
                    uid = pwd.getpwnam(self.user).pw_uid
                except KeyError:
                    log.error("No such user: %s", self.user)
                else:
                    os.setuid(uid)
                    log.debug(
                        "switched user for remote process to %s(%s)", self.user, uid
                    )
            if self.group is not None:
                try:
                    gid = grp.getgrnam(self.group).gr_gid
                except KeyError:
                    log.error("No such group: %s", self.group)
                else:
                    os.setgid(gid)
                    log.debug(
                        "switched group for remote process to %s(%s)", self.group, gid
                    )
            if not argv:
                argv = [os.environ["SHELL"]]
            os.execlp(argv[0], *argv)
            # Previous command replaces the process
            return

        self._init_fd()
        try:
            self._copy()
        except (IOError, OSError):
            pass

        os.close(master_fd)
        self.master_fd = None

    def _init_fd(self):
        """
        Called once when the pty is first set up.
        """
        self._set_pty_size()

    def _signal_winch(self, signum, frame):
        """
        Signal handler for SIGWINCH - window size has changed.
        """
        self._set_pty_size()

    def _set_pty_size(self):
        """
        Sets the window size of the child pty based on the window size of our own controlling terminal.
        """
        assert self.master_fd is not None

        # Get the terminal size of the real terminal, set it on the pseudoterminal.
        rows, cols = self.size
        buf = array.array("h", [cols, rows, 0, 0])
        # fcntl.ioctl(pty.STDOUT_FILENO, termios.TIOCGWINSZ, buf, True)
        fcntl.ioctl(self.master_fd, termios.TIOCSWINSZ, buf)

    def _copy(self):
        """
        Main select loop. Passes all data to self.master_read() or self.stdin_read().
        """
        assert self.master_fd is not None
        master_fd = self.master_fd

        poll = select.poll()
        readable_events = select.POLLIN | select.POLLPRI  # Data in , priority data in
        error_events = select.POLLERR | select.POLLHUP  #  Error or hang up
        events = readable_events | error_events
        poll.register(master_fd, events)

        reading = True
        while reading:
            try:
                poll_result = poll.poll(5 * 1000)
            except Exception as error:
                log.warning("error in proxy.py poll.poll; %s", error)
                break
            for _file_descriptor, event_mask in poll_result:
                if event_mask & readable_events:
                    data = os.read(master_fd, 1024 * 1024)
                    self.master_read(data)
                if event_mask & error_events:
                    reading = False
                    break

    def resize_terminal(self, size):
        """Resize terminal to [COLUMNS, LINES]"""
        self.size = size
        if self.master_fd is not None:
            self._set_pty_size()

    def write_stdout(self, data):
        """
        Writes to stdout as if the child process had written the data.
        """

        # os.write(pty.STDOUT_FILENO, data)

    def write_master(self, data):
        """
        Writes to the child process from its controlling terminal.
        """
        master_fd = self.master_fd
        assert master_fd is not None
        while data:
            n = os.write(master_fd, data)
            data = data[n:]

    def master_read(self, data):
        """
        Called when there is data to be sent from the child process back to the user.
        """

        self.write_stdout(data)

    def stdin_read(self, data):
        """
        Called when there is data to be sent from the user/controlling terminal down to the child process.
        """
        self.write_master(data)


if __name__ == "__main__":
    i = Interceptor()
    i.write_stdout("\nThe dream has begun.\n")
    i.spawn(sys.argv[1:])
    i.write_stdout("\nThe dream is (probably) over.\n")



from __future__ import print_function
from __future__ import unicode_literals

from os import environ


def get_environ_int(name, default):
    """Get an integer from the environment.

    Args:
        name (str): environment variable name
        default (int): Default if env var doesn't exist or not an integer

    Returns:
        int: Integer value of env var.
    """
    try:
        value = int(environ.get(name, default))
    except ValueError:
        return default
    return value


CONF_PATH = "/etc/dataplicity/dataplicity.conf"
SERVER_URL = environ.get("DATAPLICITY_API_URL", "https://api.dataplicity.com")
M2M_URL = environ.get("DATAPLICITY_M2M_URL", "wss://m2m.dataplicity.com/m2m/")
M2M_FEATURES = {"scan"}
SERIAL_LOCATION = "/opt/dataplicity/tuxtunnel/serial"
AUTH_LOCATION = "/opt/dataplicity/tuxtunnel/auth"
REMOTE_DIRECTORY_LOCATION = "/home/dataplicity/remote"


# Client will reconnect if the server hasn't responded in this time
MAX_TIME_SINCE_LAST_PACKET = 100.0  # seconds or None

# Number of bytes to read at a time, when copying date over the network
# TODO: Replace this with a sensible chunk size once we identify the
# issue with ssh over Porthole
CHUNK_SIZE = 1024 * 1024

# Maximum number of services (port forward/commands/file etc)
LIMIT_SERVICES = get_environ_int("DATAPLICITY_LIMIT_SERVICES", 500)

# Maximum number of terminals (separate pool from services)
LIMIT_TERMINALS = get_environ_int("DATAPLICITY_LIMIT_TERMINALS", 100)

# Server busy HTTP response
SERVER_BUSY = b"""HTTP/1.1 503 Device Busy\r\n\r\n
<h1>503 - Server busy</h1>

<p>The device is under heavy load and could not return a response.</p>

<p>Try increasing <tt>DATAPLICITY_LIMIT_SERVICES</tt></p>
"""



from __future__ import print_function
from __future__ import unicode_literals

import json
import logging

from .compat import urlopen, text_type


log = logging.getLogger("agent")


class ProtocolError(Exception):
    """Errors where the server didn't return the correct response"""


class ServerUnreachableError(Exception):
    """Can't reach JSONRPC server"""

    def __init__(self, url, original):
        self.url = url
        self.original = original
        _error_format = "unable to contact JSONRPC server '{url}' ({original})"
        _error = _error_format.format(url=url, original=original)
        super(ServerUnreachableError, self).__init__(_error)


class InvalidResponseError(ProtocolError):
    """Probably not JSON in response"""


class JSONRPCError(Exception):
    """Base class for exceptions returned from the server"""

    def __init__(self, method, code, data, message):
        self.method = method
        self.code = code
        self.data = data
        self.message = message
        super(JSONRPCError, self).__init__(message)


class RemoteError(JSONRPCError):
    """One of the generic error types defined in ErrorCode"""


class RemoteMethodError(JSONRPCError):
    """An error returned from the server"""


class ErrorCode(object):
    """Enumeration of JSONRPC error codes"""

    parse_error = -32700
    invalid_request = -32600
    method_not_found = -32601
    invalid_params = -32602
    internal_error = -32603

    to_str = {
        -32700: "Parse error",
        -32600: "Invalid Request",
        -32601: "Method not found",
        -32602: "Invalid params",
        -32603: "Internal error",
    }


class Batch(object):
    """An object that stores a batch of rpc calls

    May be used as a context manager

        with client.batch() as batch:
            batch.call("method1", foo="bar")
            batch.call("method2", foo="baz")

    Method call results are stored in `results` which maps the call id on to results.
    Errors are contained in `errors` which maps the call id on to a dictionary with error code / message

    """

    def __init__(self, client):
        self.client = client
        self.calls = []
        self.sent = False
        self.responses = None
        self.errors = None
        self.results = None
        self.ids_used = set()
        self.methods = {}
        self._abandoned = False
        super(Batch, self).__init__()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if not exc_type and not self._abandoned:
            self.send()

    def call(self, method, **params):
        """Add a call to the batch, using a default id."""
        call = {
            "jsonrpc": "2.0",
            "method": method,
            "params": params,
            "id": self.client.new_call_id(),
        }
        self.calls.append(call)
        self.methods[call["id"]] = method

    def call_with_id(self, call_id, method, **params):
        """Add a call to the batch with a supplied id."""
        if call_id in self.ids_used:
            raise ValueError("duplicate call id in batch")
        call = {"jsonrpc": "2.0", "method": method, "params": params, "id": call_id}
        self.calls.append(call)
        self.ids_used.add(call_id)
        self.methods[call["id"]] = method

    def notify(self, method, **params):
        call = {
            "jsonrpc": "2.0",
            "method": method,
            "params": params,
        }
        self.calls.append(call)

    def send(self):
        response_json = self.client._send(self.calls)
        self.sent = True
        response = json.loads(response_json)

        if not isinstance(response, list):
            raise ProtocolError("Expected a list of response from the server")

        self.responses = [
            (r.get("error", None), r.get("result", None)) for r in response
        ]
        self.errors = {r["id"]: r["error"] for r in response if "error" in r}
        self.results = {r["id"]: r for r in response if "id" in r and "error" not in r}

    def get_result(self, call_id, default=Ellipsis):
        """Get a result from the batch, potentially raising rpc errors."""
        if call_id in self.results:
            return self.results[call_id].get("result", None)
        elif call_id in self.errors:
            if default is Ellipsis:
                self.client._handle_error(self.methods[call_id], self.errors[call_id])
            else:
                return default
        else:
            raise KeyError("No such call_id in response")

    def check(self, *call_ids):
        """Check call IDs for exceptions, discard results."""
        for call_id in call_ids:
            self.get_result(call_id)

    def abandon(self, reason=None):
        """Mark batch as 'abandoned' (will not be sent)."""
        self._abandoned = True


class JSONRPC(object):
    """A client for a JSONRPC server."""

    unknown_error_msg = "the server did not supply further information"

    def __init__(self, url):
        self.url = url
        self.call_id = 1

    def new_call_id(self):
        self.call_id += 1
        return self.call_id

    def _send(self, call):
        call_json = json.dumps(call)
        # Py2 returns bytes, Py3 returns unicode str
        if isinstance(call_json, text_type):
            call_json = call_json.encode("utf-8")
        log.debug("JSONRPC request %i Kb", len(call_json) // 1000)
        url_file = None
        try:
            try:
                url_file = urlopen(self.url, call_json)
                response_json = url_file.read().decode("utf-8")
            finally:
                if url_file is not None:
                    url_file.close()
        except Exception as e:
            raise ServerUnreachableError(self.url, e)
        log.debug(response_json[:1000])

        return response_json

    def call(self, method, **params):
        """Call a remote method."""
        call_id = self.new_call_id()
        call = {"jsonrpc": "2.0", "method": method, "params": params, "id": call_id}
        response_json = self._send(call)
        try:
            response = json.loads(response_json)
        except:
            log.error("unable to decode %s", repr(response_json)[:100])
            raise InvalidResponseError("unable to decode response as JSON")

        if "jsonrpc" not in response or "id" not in response:
            raise ProtocolError("Invalid response from server")

        if response["jsonrpc"] != "2.0":
            raise ProtocolError("Client only understands JSONRPC v2.0")

        if response["id"] != call_id:
            raise ProtocolError(
                "Invalid response from the server, 'id' field does not match"
            )

        if "error" in response:
            error = response["error"]
            self._handle_error(method, error)

        return response.get("result", None)

    def notify(self, method, **params):
        """Send a notification to the server."""
        notify = {"jsonrpc": "2.0", "method": method, "params": params}
        self._send(notify)

    def batch(self):
        """Create a batch object that can be used to send multiple calls / notifications."""
        return Batch(self)

    def _handle_error(self, method, error):
        code = error.get("code")
        if code in ErrorCode.to_str:
            raise RemoteError(
                method,
                code,
                error.get("data", None),
                error.get("message", ErrorCode.to_str[code]),
            )
        raise RemoteMethodError(
            method,
            code,
            error.get("data", None),
            error.get("message", self.unknown_error_msg),
        )



"""
Raspberry Pi specific tools.

"""

from __future__ import unicode_literals


def get_machine_revision():
    """
    Get the machine revision.

    A return value of None indicates an error (possibly because its
    not running on an RPI).
    Otherwise the return value is the revision code.

    """
    revision_code = None
    try:
        with open("/proc/cpuinfo", "rb") as fp:
            for line in fp:
                if b":" not in line:
                    continue
                key, _, value = line.partition(b":")
                key = key.strip().lower()
                value = value.strip().lower()
                if key == b"revision":
                    revision_code = value
                    break
    except IOError:
        return None
    else:
        if isinstance(revision_code, bytes):
            return revision_code.decode("utf-8")
        return revision_code



"""
Port forwarding client.

Reads and writes to a socket, proxied over m2m.

"""

from __future__ import print_function
from __future__ import unicode_literals

from time import time
import logging
import select
import socket
import threading
import weakref

from .constants import CHUNK_SIZE, SERVER_BUSY


log = logging.getLogger("pf")


class Connection(threading.Thread):
    """Handles a single remote controlled TCP/IP connection."""

    def __init__(self, limiter, close_event, channel, host_port):
        """Initialize the connection, set up callbacks."""
        super(Connection, self).__init__()
        self.limiter = limiter
        self._close_event = close_event
        self.channel = channel
        self.host_port = host_port

        self._lock = threading.RLock()
        self._start_time = time()
        self.socket = None
        self.read_buffer = []  # For data received before we connected

        self.channel.set_callbacks(
            self.on_channel_data, self.on_channel_close, self.on_channel_control
        )

    @property
    def close_event(self):
        """Get a threading.Event object."""
        return self._close_event

    def run(self):
        """Run the main loop, and decrement limiter."""
        try:
            self._run()
        finally:
            # Ensure decrement is called if _run throws an exception
            self.limiter.decrement()

    def _run(self):
        """
        Main loop, connects to local server, reads data, and writes it to an
        m2m channel.
        """
        bytes_written = 0
        readable_events = select.POLLIN | select.POLLPRI  # Data in , priority data in
        error_events = select.POLLERR | select.POLLHUP  #  Error or hang up
        events = readable_events | error_events
        poll = select.poll()
        try:
            # Connect to remote host
            if not self._connect():
                return

            poll.register(self.socket.fileno(), events)

            log.debug("entered recv loop")
            # Read all the data we can and write it to the channel
            while not self.close_event.is_set():
                # Block for a period of time until the socket becomes readable,
                # or there is an error
                try:
                    poll_result = poll.poll(5 * 1000)
                except Exception as error:
                    # For paranoia only.
                    log.warning("error in portforward.py poll.poll; %s", error)
                    break
                if self.channel.is_closed:
                    break
                for _file_descriptor, event_mask in poll_result:
                    if event_mask & readable_events:
                        try:
                            # Reads *up to* BUFFER_SIZE bytes
                            data = self.socket.recv(CHUNK_SIZE)
                        except Exception:
                            log.exception("error in recv")
                            break
                        else:
                            if data:
                                self.channel.write(data)
                                bytes_written += len(data)
                            else:
                                # No data means the socket has been closed
                                break
                    if event_mask & error_events:
                        # Socket has been closed in another thread, possibly due to
                        # m2m channel closing
                        self.close_event.set()
        finally:

            speed = bytes_written / 1024.0 / (time() - self._start_time)
            log.debug("left recv loop (read %s bytes) %0.1fKB/s ", bytes_written, speed)
            # These close methods are a null operation if the objects are
            # already closed
            self.channel.close()
            self._shutdown_read()

    def _shutdown_read(self):
        """Shutdown reading."""
        with self._lock:
            if self.socket:
                try:
                    self.socket.shutdown(socket.SHUT_RD)
                except Exception:
                    pass

    def _shutdown_write(self):
        """Shutdown writing."""
        with self._lock:
            if self.socket:
                try:
                    self.socket.shutdown(socket.SHUT_WR)
                except Exception:
                    pass

    def _close_socket(self):
        """Shutdown the socket."""
        with self._lock:
            if self.socket is not None:
                try:
                    self.socket.shutdown(socket.SHUT_RDWR)
                except Exception:
                    pass
                try:
                    self.socket.close()
                except Exception:
                    log.exception("error closing socket")

    def connect(self):
        """Start thread, and connect to local server."""
        # Connect may block, so do it in a thread

        try:
            with self.limiter():
                self.start()
        except Exception as error:
            log.warning("unable to start portforard thread; %r", error)
            self.channel.write(SERVER_BUSY)
            self.channel.close()

    def _connect(self):
        """Connect to a local server, return True on success."""
        _socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # No Nagle since we are going for as close to realtime as possible
        _socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        # Set the timeout for initial connect, as default is too high
        _socket.settimeout(5.0)

        log.debug("connecting to %s:%d", *self.host_port)
        try:
            _socket.connect(self.host_port)
        except socket.timeout:
            log.error("timed out connecting to server")
            return False
        except IOError as e:
            log.error("IO Error when connecting, %s", e)
            return False
        except Exception:
            log.exception("error connecting")
            return False
        else:
            log.debug("connected to %s:%d", *self.host_port)
            self.socket = _socket
            self._flush_buffer()
            return True

    def on_channel_data(self, data):
        """Called by m2m channel."""
        with self._lock:
            self.read_buffer.append(data)
            self._flush_buffer()

    def _flush_buffer(self):
        with self._lock:
            if self.socket is not None:
                try:
                    for chunk in self.read_buffer:
                        self.socket.sendall(chunk)
                finally:
                    del self.read_buffer[:]

    def on_channel_close(self):
        """Called when the channel has been closed."""
        log.debug("channel close")
        with self._lock:
            # Shut down the socket
            # This will cause an exceptional condition in the select loop,
            # which will subsequently exit cleanly
            self._flush_buffer()
            self._shutdown_write()

    def on_channel_control(self, data):
        """Called when the remote end sends a control packet (currently not used)."""
        log.debug("channel control %r", data)


class Service(object):
    """A service defines a host and port to forward."""

    def __init__(self, manager, name, port, host="127.0.0.1"):
        self._manager = weakref.ref(manager)
        self.name = name
        self.port = port
        self.host = host
        self.m2m_port = None
        self._connect_index = 0
        self._lock = threading.RLock()

    def __repr__(self):
        """Some useful info re the service."""
        return "<service {}:{} '{}'>".format(self.host, self.port, self.name)

    @property
    def manager(self):
        """Get the manager from weakref."""
        return self._manager()

    @property
    def m2m(self):
        """Get the M2M interface."""
        return self.manager.m2m

    @property
    def close_event(self):
        """The one close event to rule them all."""
        return self.manager.close_event

    @property
    def host_port(self):
        """A tuple of (host, port) as a convenience for socket.connect."""
        return (self.host, self.port)

    def connect(self, limiter, port_no):
        """Add a new connection."""
        self.m2m_port = port_no
        channel = self.m2m.m2m_client.get_channel(port_no)
        log.debug("new %r connection on port %s", self, port_no)
        try:
            with limiter():
                with self._lock:
                    connection = Connection(
                        limiter, self.close_event, channel, self.host_port
                    )
                connection.start()
        except Exception as error:
            channel.write(SERVER_BUSY)
            channel.close()
            log.warning("failed to connect %r; %s", self, error)


class PortForwardManager(object):
    """Managed port forwarded services."""

    def __init__(self, client):
        self._client = weakref.ref(client)
        self._services = {}
        self._ports = {}
        self._close_event = threading.Event()

    @property
    def client(self):
        return self._client()

    @property
    def m2m(self):
        return self.client.m2m if self.client else None

    @classmethod
    def init(cls, client):
        manager = cls(client)
        manager.add_service("web", 80)
        manager.add_service("ext", 81)
        manager.add_service("extalt", 8000)
        manager.add_service("alt", 8080)
        return manager

    @property
    def close_event(self):
        return self._close_event

    def on_client_close(self):
        """M2M client closed."""
        log.debug("m2m exited")

    def get_service_on_port(self, port):
        """Get the service on a numbered port."""
        service_name = self._ports.get(port)
        if service_name is None:
            return None
        return self._services[service_name]

    def get_service(self, service, default=None):
        """Get a named service."""
        return self._services.get(service, default)

    def add_service(self, name, port, host="127.0.0.1"):
        """Add a service to be exposed."""
        service = Service(self, name, port, host=host)
        self._services[name] = service
        self._ports[port] = name
        log.debug("added port forward service '%s' on port %s", name, port)

    def open_service(self, limiter, service, route):
        log.debug("opening service %s on %r", service, route)
        node1, port1, node2, port2 = route
        self.open(limiter, port2, service)

    def open(self, limiter, m2m_port, service=None, port=None):
        """Open a port forward service."""
        if service is None and port is None:
            raise ValueError("one of service or port is required")
        if port is not None:
            service = self.get_service_on_port(port or 80)
        elif service is not None:
            service = self.get_service(service)
        if service is None:
            return
        service.connect(limiter, m2m_port)

    def redirect_port(self, limiter, m2m_port, device_port):
        # we need to store the reference to the Service somewhere so that
        # when the Connection starts in thread it wouldn't loose the value
        # of service variable. However, we have to remember that there may
        # be numerous connections to the same local port.
        # for instance, one could be ssh'ed into a machine twice, so we
        # shan't confuse these two connections.
        # therefore, an easy way is to store these in a dict, so that the
        # lookup would be quick
        #
        channel = self.m2m.m2m_client.get_channel(m2m_port)
        try:
            with limiter():
                Connection(
                    limiter,
                    close_event=self.close_event,
                    channel=channel,
                    host_port=("127.0.0.1", device_port),
                ).start()
        except Exception as error:
            log.warning("unable to start portforard thread; %r", error)
            channel.write(SERVER_BUSY)
            channel.close()



"""Manages M2M connections."""

from __future__ import print_function
from __future__ import unicode_literals

import logging
import subprocess
import threading

from . import constants
from .compat import PY3
from .limiter import Limiter
from .m2m import WSClient, EchoService
from .m2m.fileservice import FileService
from .m2m.commandservice import CommandService
from .m2m.remoteprocess import RemoteProcess


log = logging.getLogger("m2m")


class Terminal(object):
    """Configured terminal information."""

    def __init__(self, name, command, user=None, group=None):
        self.name = name
        self.command = command
        self.user = user
        self.group = group
        self.processes = []

    def __repr__(self):
        return "<terminal '{}' command='{}'>".format(self.name, self.command)

    def _prune_closed(self):
        """Remove closed processes."""
        self.processes[:] = [
            process for process in self.processes if not process.is_closed
        ]

    def launch(self, limiter, channel, size=None):
        """Launch a terminal instance."""

        if size is None:
            size = [80, 24]
        self._prune_closed()
        log.debug("opening terminal %s", self.name)
        remote_process = None
        try:
            remote_process = RemoteProcess(
                limiter,
                self.command,
                channel,
                user=self.user,
                group=self.group,
                size=size,
            )
        except Exception:
            log.exception("error launching terminal process '%s'", self.command)
            if remote_process is not None:
                try:
                    remote_process.close()
                except:
                    pass
        else:
            try:
                with limiter():
                    process_thread = threading.Thread(target=remote_process.run)
                    process_thread.start()
                    log.info("launched remote process %r over %r", self, channel)
            except Exception as error:
                log.info("unable to launch remote process; %s", error)
                channel.write(b"Failed to launch remote process\n")
                channel.close()
            else:
                self.processes.append(remote_process)

    def close(self):
        self._prune_closed()
        for process in self.processes:
            log.debug("closing %r", self)
            try:
                if not process.is_closed:
                    log.debug("closing %r", self)
                    process.close()
            except:
                log.exception("error closing %s", process)
        del self.processes[:]


class M2MManager(object):
    """Manages M2M Services."""

    def __init__(self, client, url, remote_directory, identity=None):
        self.client = client
        self.url = url
        self.identity = identity
        self.terminals = {}
        self.notified_identity = None
        self.m2m_client = WSClient(self, url, remote_directory)
        self.services_limiter = Limiter("services", constants.LIMIT_SERVICES)
        self.terminals_limiter = Limiter("terminals", constants.LIMIT_TERMINALS)

    @classmethod
    def init(cls, client, remote_directory, m2m_url=None):
        """Set up the m2m manager for Dataplicity."""
        url = m2m_url or constants.M2M_URL
        manager = cls(client, url, remote_directory)
        manager.m2m_client.start()
        manager.add_terminal("shell", "bash -i")
        return manager

    def restart_agent(self):
        """Restart the agent."""
        self.client.exit()

    def on_client_close(self):
        """Client is closing, shutdown all terminals."""
        for terminal in self.terminals.values():
            terminal.close()

    def set_identity(self, identity):
        """Set the m2m identity, and also notifies the dataplicity server if required."""
        self.identity = identity
        if identity and identity != self.notified_identity:
            log.info("m2m identity changed (%s)", identity.decode("utf-8", "replace"))
            self.notified_identity = self.client.set_m2m_identity(identity)

    def on_sync(self, batch):
        """Called by sync, so it can inject commands in to the batch request."""
        # Send the m2m identity on every sync
        # This shouldn't be necessary, but could mitigate any screw ups server side
        # NOTE: not currently called in agent
        if self.identity:
            log.debug("syncing m2m identity (%s)", self.identity)
            batch.notify("m2m.associate", identity=self.identity)

    def close(self):
        log.debug("m2m manager close")
        if self.m2m_client is not None:
            self.m2m_client.close()

    def add_terminal(self, name, remote_process, user=None, group=None):
        """Add a terminal for a remote process."""
        log.debug("adding terminal '%s' %s", name, remote_process)
        self.terminals[name] = Terminal(name, remote_process, user=user, group=group)

    def get_terminal(self, name):
        """Get a named terminal."""
        return self.terminals.get(name, None)

    def on_instruction(self, sender, data):
        """Instructions sent by privileged client."""
        log.debug("instruction %r from %s", data, sender)

        if PY3:

            def decode_string(value):
                return (
                    value.decode("utf-8", "ignore")
                    if isinstance(value, bytes)
                    else value
                )

            data = dict(
                (key.decode("utf-8", "ignore"), decode_string(value))
                for key, value in data.items()
            )
        action = data["action"]
        if action == "sync":
            self.client.sync()
        elif action == "open-terminal":
            port = data["port"]
            terminal_name = data["name"]
            size = data.get("size", None)
            self.open_terminal(terminal_name, port, size=size)
        elif action == "open-echo":
            port = data["port"]
            self.open_echo_service(port)
        elif action == "open-portforward":
            service = data["service"]
            route = data["route"]
            self.open_portforward(service, route)
        elif action == "open-portredirect":
            device_port = data["device_port"]
            m2m_port = data["m2m_port"]
            self.client.port_forward.redirect_port(m2m_port, device_port)
        elif action == "reboot-device":
            self.reboot()
        elif action == "read-file":
            self.open_file_service(data["port"], data["path"])
        elif action == "run-command":
            self.open_command_service(data["port"], data["command"])
        elif action == "scan-directory":
            self.client.directory_scanner.perform_scan()
        # Unrecognized instructions are ignored

    def open_terminal(self, name, port, size=None):
        """Open a new terminal."""
        terminal = self.get_terminal(name)
        if terminal is None:
            log.warning("no terminal called '%s'", name)
            return
        terminal.launch(
            self.terminals_limiter, self.m2m_client.get_channel(port), size=size
        )

    def open_echo_service(self, port):
        """Open an echo service (ping)."""
        log.debug("opening echo service on m2m port %s", port)
        # Doesn't use threads, so doesn't need limiter
        EchoService(self.m2m_client.get_channel(port))

    def open_portforward(self, service, route):
        """Open a port forward service."""
        self.client.port_forward.open_service(self.services_limiter, service, route)

    def reboot(self):
        """Initiate a reboot."""
        # TODO: consider initiating a graceful shutdown of dpcore that ends in a rebooot
        command = "/usr/bin/sudo /sbin/reboot"
        log.debug("rebooting!")
        # Why not subprocess.call? Because it will block this process and prevent graceful exit!
        pid = subprocess.Popen(command.split()).pid
        log.debug("opened reboot process %s", pid)

    def open_file_service(self, port, path):
        """Open a file service, to send a file over a port."""
        channel = self.m2m_client.get_channel(port)
        FileService(self.services_limiter, channel, path)

    def open_command_service(self, port, command):
        """Open a service that runs a command and sends the stdout over m2m."""
        channel = self.m2m_client.get_channel(port)
        CommandService(self.services_limiter, channel, command)



from __future__ import unicode_literals
from __future__ import print_function

import logging
import platform
from threading import Event, Lock
import os
import random
import sys
import time

from ._version import __version__
from . import constants
from . import device_meta
from . import jsonrpc
from .clockcheck import ClockCheckThread
from .directory_scanner import DirectoryScanner
from .disk_tools import disk_usage
from .m2mmanager import M2MManager
from .portforward import PortForwardManager
from .remote_directory import RemoteDirectory
from .tags import get_tag_list, TagError
import six

log = logging.getLogger("agent")


class Client(object):
    """Dataplicity client."""

    def __init__(
        self,
        rpc_url=None,  # type: str
        m2m_url=None,  # type: str
        serial=None,  # type: str
        auth_token=None,  # type: str
        remote_directory_path=None,  # type: str
    ):
        # type: (...) -> None
        self.rpc_url = rpc_url or constants.SERVER_URL
        self.m2m_url = m2m_url or constants.M2M_URL

        if "?" not in self.m2m_url:
            self.m2m_url += "?features=" + ",".join(constants.M2M_FEATURES)

        self.auth_token = auth_token
        self.serial = serial
        self.remote_directory_path = remote_directory_path

        self._sync_lock = Lock()
        self._sent_meta = False
        self.exit_event = Event()
        self._init()
        self._tag_list = None

    @classmethod
    def _read(cls, path):
        """Read contents of a file, strip whitespace."""
        path = os.path.expanduser(path)
        with open(path, "rt") as fh:
            data = fh.read().strip()
        return data

    def _init(self):
        try:
            log.info("dataplicity %s", __version__)
            log.info("python executable=%s", sys.executable)
            log.info("python version=%s", sys.version.replace("\n", " "))
            log.info("uname=%s", " ".join(platform.uname()))

            self.remote = jsonrpc.JSONRPC(self.rpc_url)
            self.serial = self.serial or self._read(constants.SERIAL_LOCATION)
            self.auth_token = self.auth_token or self._read(constants.AUTH_LOCATION)
            self.remote_directory_path = (
                self.remote_directory_path or constants.REMOTE_DIRECTORY_LOCATION
            )
            log.info("remote_directory=%s", self.remote_directory_path)

            self.poll_rate_seconds = 60
            self.disk_poll_rate_seconds = 60 * 60
            self.next_disk_poll_time = time.time()

            log.info("m2m=%s", self.m2m_url)
            log.info("api=%s", self.rpc_url)
            log.info("serial=%s", self.serial)
            log.info("poll=%s", self.poll_rate_seconds)

            self.directory_scanner = DirectoryScanner(self.remote_directory_path)
            self.remote_directory = RemoteDirectory(
                self.remote_directory_path, self.directory_scanner
            )
            self.m2m = M2MManager.init(
                self, self.remote_directory, m2m_url=self.m2m_url
            )
            self.port_forward = PortForwardManager.init(self)

        except Exception:
            log.exception("failed to initialize client")
            raise

    def run_forever(self):
        """Run the client "forever"."""
        clock_check_thread = ClockCheckThread()
        clock_check_thread.start()

        try:
            self.poll()
            while not self.exit_event.wait(self.poll_rate_seconds):
                self.poll()
        except SystemExit:
            log.debug("exit requested")
            return
        except KeyboardInterrupt:
            log.debug("user exit")
            return
        finally:
            clock_check_thread.running = False
            clock_check_thread.join()
            log.debug("closing")
            self.close()
            log.debug("goodbye")

    def exit(self):
        """Exit the agent."""
        self.exit_event.set()

    def disk_poll(self):
        now = time.time()

        if now >= self.next_disk_poll_time:
            self.next_disk_poll_time = now + self.disk_poll_rate_seconds
            disk_space = disk_usage("/")

            with self.remote.batch() as batch:
                batch.call_with_id(
                    "authenticate_result",
                    "device.check_auth",
                    device_class="tuxtunnel",
                    serial=self.serial,
                    auth_token=self.auth_token,
                )
                batch.call_with_id(
                    "set_disk_space_result",
                    "device.set_disk_space",
                    disk_capacity=disk_space.total,
                    disk_used=disk_space.used,
                )

    def tag_poll(self):
        """Gets the tag list for get_tag_list() and sends to the server"""
        try:
            tag_list = get_tag_list()
        except TagError:
            return

        try:
            if tag_list != self._tag_list:
                log.debug("new machine tags: %r", tag_list)
                with self.remote.batch() as batch:
                    batch.call_with_id(
                        "authenticate_result",
                        "device.check_auth",
                        device_class="tuxtunnel",
                        serial=self.serial,
                        auth_token=self.auth_token,
                    )
                    batch.call_with_id(
                        "set_machine_defined_tags_result",
                        "device.set_machine_defined_tags",
                        tag_list=tag_list,
                    )
                batch.get_result("set_machine_defined_tags_result")
        except jsonrpc.JSONRPCError as error:
            log.error(
                'unable to set tag list ("%s"=%s, "%s")',
                error.method,
                error.code,
                error.message,
            )
            return None
        except jsonrpc.ServerUnreachableError as error:
            log.debug("set tag list failed: %s", error)
            return None
        except Exception as error:
            log.error("set tag list failed: %s", error)
            return None
        else:
            # Success! Set cached tag list
            self._tag_list = tag_list

    def poll(self):
        """Called at regular intervals."""
        t = time.time()
        log.debug("poll t=%.02fs", t)
        try:
            self.disk_poll()
        except Exception as e:
            log.error("disk poll failed %s", e)

        try:
            self.tag_poll()
        except Exception as error:
            log.error("tag poll failed %s", error)

        self.sync()

    def close(self):
        """Perform shutdown."""
        pass

    @classmethod
    def make_sync_id(cls):
        """Make a random sync ID."""
        sync_id = "".join(
            random.choice("abcdefghijklmnopqrstuvwxyz") for _ in six.moves.xrange(12)
        )
        return sync_id

    def sync(self):
        """Sync with server."""
        try:
            with self._sync_lock:
                self._sync()
        except Exception as e:
            log.error("sync failed %s", e)

    def _sync(self):
        """Perform sync."""
        # Syncing is a much simpler process in Dataplicity agent,
        # than previous versions.
        start = time.time()
        sync_id = self.make_sync_id()
        try:
            if not self._sent_meta:
                with self.remote.batch() as batch:
                    batch.call_with_id(
                        "authenticate_result",
                        "device.check_auth",
                        device_class="tuxtunnel",
                        serial=self.serial,
                        auth_token=self.auth_token,
                        sync_id=sync_id,
                    )
                    self._sync_meta(batch)
                batch.get_result("authenticate_result")
                self._check_meta(batch)

        finally:
            elapsed = time.time() - start
            log.debug("sync complete %0.2fs", elapsed)

    def _sync_meta(self, batch):
        """Sync meta information regarding host device."""
        try:
            meta = device_meta.get_meta()
            log.debug("syncing meta %r", meta)
        except:
            log.exception("error getting meta")
        else:
            batch.call_with_id(
                "set_agent_version_result",
                "device.set_agent_version",
                agent_version=meta["agent_version"],
            )
            batch.call_with_id(
                "set_machine_revision_result",
                "device.set_machine_revision",
                revision_code=meta["machine_revision"],
            )
            batch.call_with_id(
                "set_os_version_result",
                "device.set_os_version",
                os_version=meta["os_version"],
            )
            batch.call_with_id(
                "set_uname_result", "device.set_uname", uname=meta["uname"]
            )
            batch.call_with_id(
                "set_ip_addresses_result",
                "device.set_ip_addresses",
                ip_list=meta["ip_list"],
            )

    def _check_meta(self, batch):
        """Check previously sent meta information."""
        log.debug("checking meta")
        if self._sent_meta:
            log.debug("meta was previously sent")
            return
        try:
            batch.check(
                "set_agent_version_result",
                "set_machine_revision_result",
                "set_os_version_result",
                "set_uname_result",
                "set_ip_addresses_result",
            )
        except Exception as e:
            log.warning("failed to set device meta (%s)", e)
        else:
            # Success! Don't send again.
            self._sent_meta = True
            log.debug("sent meta")

    def set_m2m_identity(self, identity):
        """
        Tell the server of our m2m identity, return the identity if it was set,
        or None if it could not be set.

        """
        if self.auth_token is None:
            return None

        try:
            log.debug(
                "notifying server (%s) of m2m identity (%s)",
                self.remote.url,
                identity or "<None>",
            )
            with self.remote.batch() as batch:
                batch.call_with_id(
                    "authenticate_result",
                    "device.check_auth",
                    device_class="tuxtunnel",
                    serial=self.serial,
                    auth_token=self.auth_token,
                )
                batch.call_with_id(
                    "associate_result",
                    "m2m.associate",
                    identity=identity.decode("utf-8") or "",
                )
            # These methods may potentially throw JSONRPCErrors
            batch.get_result("authenticate_result")
            batch.get_result("associate_result")
        except jsonrpc.JSONRPCError as e:
            log.error(
                'unable to associate m2m identity ("%s"=%s, "%s")',
                e.method,
                e.code,
                e.message,
            )
            return None
        except jsonrpc.ServerUnreachableError as e:
            log.debug("set m2m identity failed, %s", e)
            return None
        except Exception as error:
            log.error("unable to set m2m identity: %s", error, exc_info=False)
            return None
        else:
            # If we made it here the server has acknowledged it received the identity
            log.debug("server received m2m identity %s", identity)
            return identity



from __future__ import unicode_literals
from __future__ import print_function

import argparse
import logging
import logging.config
import sys

from . import __version__
from . import subcommand
from .client import Client
from .subcommands import run, version

log = logging.getLogger("app")

# Map log levels on to integer values
_logging_level_names = {
    "NOTSET": 0,
    "DEBUG": 10,
    "INFO": 20,
    "WARN": 30,
    "WARNING": 30,
    "ERROR": 40,
    "CRITICAL": 50,
}


class App(object):
    """Dataplicity Agent command line interface."""

    def __init__(self):
        self.subcommands = {
            name: cls(self) for name, cls in subcommand.registry.items()
        }

    def _make_arg_parser(self):
        """Make an argument parse object."""
        parser = argparse.ArgumentParser("dataplicity", description=self.__doc__)

        _version = "dataplicity agent v{}".format(__version__)
        parser.add_argument(
            "-v",
            "--version",
            action="version",
            version=_version,
            help="Display version and exit",
        )
        parser.add_argument(
            "--log-level",
            metavar="LEVEL",
            default="INFO",
            help="Set log level (INFO or WARNING or ERROR or DEBUG)",
        )
        parser.add_argument(
            "--log-file", metavar="PATH", default=None, help="Set log file"
        )
        parser.add_argument(
            "-d",
            "--debug",
            action="store_true",
            dest="debug",
            default=False,
            help="Enables debug output",
        )
        parser.add_argument(
            "-s",
            "--server-url",
            metavar="URL",
            dest="server_url",
            default=None,
            help="URL of dataplicity.com api",
        )
        parser.add_argument(
            "-m",
            "--m2m-url",
            metavar="WS URL",
            dest="m2m_url",
            default=None,
            help="URL of m2m server (should start with ws:// or wss://",
        )
        parser.add_argument(
            "-q", "--quiet", action="store_true", default=False, help="Hide output"
        )
        parser.add_argument(
            "--serial",
            dest="serial",
            metavar="SERIAL",
            default=None,
            help="Set Dataplicity serial",
        )
        parser.add_argument(
            "--auth",
            dest="auth_token",
            metavar="KEY",
            default=None,
            help="Set Dataplicity auth token",
        )

        parser.add_argument(
            "--remote-dir",
            dest="remote_directory",
            metavar="PATH",
            default=None,
            help="Set remote directory location",
        )

        subparsers = parser.add_subparsers(
            title="available sub-commands", dest="subcommand", help="sub-command help"
        )

        for name, _subcommand in self.subcommands.items():
            subparser = subparsers.add_parser(
                name,
                help=_subcommand.help,
                description=getattr(_subcommand, "__doc__", None),
            )
            _subcommand.add_arguments(subparser)
        return parser

    def _init_logging(self):
        """Initialise logging."""
        log_format = "%(asctime)s %(name)s\t: %(message)s"
        log_level = "CRITICAL" if self.args.quiet else self.args.log_level.upper()
        try:
            log_level_no = _logging_level_names[log_level]
        except IndexError:
            self.error("invalid log level")

        if self.args.log_file:
            log_config = {
                "version": 1,
                "disable_existing_loggers": False,
                "formatters": {
                    "simple": {
                        "class": "logging.Formatter",
                        "format": log_format,
                        "datefmt": "[%d/%b/%Y %H:%M:%S]",
                    }
                },
                "handlers": {
                    "file": {
                        "level": log_level,
                        "class": "logging.handlers.RotatingFileHandler",
                        "maxBytes": 5 * 1024 * 1024,
                        "backupCount": 5,
                        "filename": self.args.log_file,
                        "formatter": "simple",
                    }
                },
                "loggers": {"": {"level": log_level, "handlers": ["file"]}},
            }
            logging.config.dictConfig(log_config)
        else:
            logging.basicConfig(
                format=log_format, datefmt="[%d/%b/%Y %H:%M:%S]", level=log_level_no
            )

    def make_client(self):
        """Make the client object."""
        client = Client(
            rpc_url=self.args.server_url,
            m2m_url=self.args.m2m_url,
            serial=self.args.serial,
            auth_token=self.args.auth_token,
            remote_directory_path=self.args.remote_directory,
        )
        return client

    def error(self, msg, code=-1):
        """Display error and exit app."""
        log.critical("app exit ({%s}) code={%s}", msg, code)
        sys.stderr.write(msg + "\n")
        sys.exit(code)

    def run(self):
        parser = self._make_arg_parser()
        args = self.args = parser.parse_args(sys.argv[1:])

        self._init_logging()
        log.debug("ready")

        if args.subcommand is None:
            parser.print_help()
            return 1

        subcommand = self.subcommands[args.subcommand]
        subcommand.args = args

        try:
            return subcommand.run() or 0
        except Exception as e:
            if self.args.debug:
                raise
            sys.stderr.write("(dataplicity {}) {}\n".format(__version__, e))
            cmd = sys.argv[0].rsplit("/", 1)[-1]
            debug_cmd = " ".join([cmd, "--debug"] + sys.argv[1:])
            sys.stderr.write("(run '{}' for a full traceback)\n".format(debug_cmd))
            return -1


def main():
    """Dataplicity Agent entry point."""
    return_code = App().run() or 0
    log.debug("exit with code %s", return_code)
    sys.exit(return_code)



"""

An M2M Service to run commands.

The stdout is sent a line at a time, until the command is finished,
and the channel is closed.

"""

from __future__ import print_function
from __future__ import unicode_literals

import logging
import os
import select
import subprocess
import threading

from lomond.errors import WebSocketError

from ..limiter import LimitReached
from ..constants import CHUNK_SIZE


log = logging.getLogger("m2m")


class CommandService(threading.Thread):
    """Runs a command and sends the stdout over m2m."""

    def __init__(self, limiter, channel, command):
        self.limiter = limiter
        self._repr = "CommandService({!r}, {!r})".format(channel, command)
        try:
            with limiter():
                super(CommandService, self).__init__(
                    args=(channel, command), target=self.run_service
                )
                self.start()
        except Exception as error:
            log.warning("unable to launch %r; %s", self, error)
            self.send_error(channel, "error", str(error))
            channel.close()

    def __repr__(self):
        return self._repr

    @classmethod
    def send_error(cls, channel, status, msg, **extra):
        """Send a control packet with an error"""
        error = {"service": "command", "type": "error", "status": status, "msg": msg}
        error.update(extra)
        channel.send_control(error)

    def run_service(self, channel, command):
        """Run the thread and log exceptions."""
        try:
            try:
                self._run_service(channel, command)
            except Exception:
                log.exception("error running %r", self)
                self.send_error(channel, "error", "error running command")
        finally:
            self.limiter.decrement()

    def _run_service(self, channel, command):
        """Run command and send stdout over m2m."""
        log.debug("%r started", self)
        process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
        )
        bytes_sent = 0
        stdout_fh = process.stdout.fileno()
        stderr_fh = process.stderr.fileno()

        readable_events = select.POLLIN | select.POLLPRI  # Data in , priority data in
        error_events = select.POLLERR | select.POLLHUP  #  Error or hang up
        events = readable_events | error_events
        poll = select.poll()
        poll.register(stdout_fh, events)
        poll.register(stderr_fh, events)
        try:
            more_data = True
            while more_data:
                try:
                    poll_result = poll.poll(0.5 * 1000)
                except Exception as error:
                    log.warning("error in commandservice.py poll.poll; %s", error)
                    break

                for _file_descriptor, event_mask in poll_result:
                    if event_mask & readable_events:
                        if _file_descriptor == stdout_fh:
                            chunk = os.read(stdout_fh, CHUNK_SIZE)
                            if not chunk:
                                more_data = False
                                log.debug("%r EOF", self)
                                break
                            channel.write(chunk)
                            bytes_sent += len(chunk)
                        else:
                            chunk = os.read(stderr_fh, CHUNK_SIZE)
                            log.debug("%r [stderr] %r", self, chunk)
                    if event_mask & error_events:
                        break
                if channel.is_closed:
                    log.debug("%r channel closed", self)
                    break

        except WebSocketError as websocket_error:
            log.warning("%r websocket error (%s)", self, websocket_error)
            # Can't send error message if websocket is fubar
        except Exception as error:
            log.exception("%r error", self)
            self.send_error(channel, "error", "error running command")
        else:
            log.info('read %s byte(s) from command "%s"', bytes_sent, command)
        finally:
            channel.send_control(
                {"service": "command", "type": "complete", "returncode": process.poll()}
            )
            channel.close()
            try:
                if process.poll() is None:
                    process.kill()
            except OSError:
                log.exception("%r failed to kill", self)



from __future__ import division
from collections import namedtuple

import os
import sys


sdiskusage = namedtuple("sdiskusage", ["total", "used", "free", "percent"])


def usage_percent(used, total, _round=None):
    """Calculate percentage usage of 'used' against 'total'."""
    try:
        ret = (used / total) * 100
    except ZeroDivisionError:
        ret = 0.0 if isinstance(used, float) or isinstance(total, float) else 0
    if _round is not None:
        return round(ret, _round)
    else:
        return ret


def disk_usage(path):
    """Return disk usage associated with path.
    Note: UNIX usually reserves 5% disk space which is not accessible
    by user. In this function "total" and "used" values reflect the
    total and used disk space whereas "free" and "percent" represent
    the "free" and "used percent" user disk space.
    Code from https://github.com/giampaolo/psutil/blob/master/psutil/_psposix.py
    """
    try:
        st = os.statvfs(path)
    except UnicodeEncodeError:
        if isinstance(path, unicode):
            # this is a bug with os.statvfs() and unicode on
            # Python 2, see:
            # - https://github.com/giampaolo/psutil/issues/416
            # - http://bugs.python.org/issue18695
            try:
                path = path.encode(sys.getfilesystemencoding())
            except UnicodeEncodeError:
                pass
            st = os.statvfs(path)
        else:
            raise

    # Total space which is only available to root (unless changed
    # at system level).
    total = st.f_blocks * st.f_frsize
    # Remaining free space usable by root.
    avail_to_root = st.f_bfree * st.f_frsize
    # Remaining free space usable by user.
    avail_to_user = st.f_bavail * st.f_frsize
    # Total space being used in general.
    used = total - avail_to_root
    # Total space which is available to user (same as 'total' but
    # for the user).
    total_user = used + avail_to_user
    # User usage percent compared to the total amount of space
    # the user can use. This number would be higher if compared
    # to root's because the user has less space (usually -5%).
    usage_percent_user = usage_percent(used, total_user, _round=1)

    # NB: the percentage is -5% than what shown by df due to
    # reserved blocks that we are currently not considering:
    # https://github.com/giampaolo/psutil/issues/829#issuecomment-223750462
    return sdiskusage(
        total=total, used=used, free=avail_to_user, percent=usage_percent_user
    )



"""A M2M service that returns data sent to it. May be used as a heart-beat."""

from __future__ import print_function
from __future__ import unicode_literals

import logging
import weakref

log = logging.getLogger("m2m")


class EchoService(object):
    """
    M2M echo service.

    Data will be sent back on the same channel.

    """

    def __init__(self, channel):
        # When the channel is closed, this object should go out of scope
        self.channel = weakref.ref(channel)
        channel.set_callbacks(on_data=self.on_data)

    def on_data(self, data):
        # Send data back
        self.channel().write(data)



from __future__ import unicode_literals

import logging
import platform

from .iptool import get_all_interfaces
from . import rpi
from ._version import __version__


log = logging.getLogger("agent")


# Cache the meta dict because it never changes
_META_CACHE = None


def get_meta():
    """Get a dict containing device meta information."""
    global _META_CACHE
    if _META_CACHE is not None:
        return _META_CACHE.copy()
    meta = {}
    meta["agent_version"] = __version__
    meta["machine_revision"] = rpi.get_machine_revision()
    meta["os_version"] = get_os_version()
    meta["uname"] = get_uname()
    meta["ip_list"] = get_ip_address_list()
    _META_CACHE = meta
    return meta


def get_uname():
    """Get uname."""
    # Preferable to running a system command
    uname = " ".join(platform.uname())
    return uname


def get_os_version():
    """Get the OS version."""
    # Linux is a fair assumption for now
    distro = " ".join(platform.linux_distribution()).strip()
    return distro


def get_ip_address_list():
    # Get the ip addresses from all the interfaces
    try:
        interfaces = get_all_interfaces()
    except Exception:
        log.exception("unable to retrieve interface information")
        # Sorry for the pokemon exception, but I don't know how
        # reliable the call is, and if it fails what it will fail with.
        # It needs some exception handling or the whole get_meta call
        # will fail
        return []
    return [i[1] for i in interfaces]



#!/usr/bin/env python

from setuptools import setup, find_packages

classifiers = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "Programming Language :: Python",
]

# http://bit.ly/2alyerp
with open("dataplicity/_version.py") as f:
    exec(f.read())

with open("README.md") as f:
    long_desc = f.read()

setup(
    name="dataplicity",
    version=__version__,
    description="Platform for connected devices",
    long_description=long_desc,
    author="WildFoundry",
    author_email="support@dataplicity.com",
    url="https://www.dataplicity.com",
    platforms=["any"],
    packages=find_packages(),
    classifiers=classifiers,
    entry_points={"console_scripts": ["dataplicity = dataplicity.app:main"]},
    install_requires=["enum34==1.1.6", "six==1.10.0", "lomond==0.3.3"],
    zip_safe=True,
)



from __future__ import unicode_literals
from __future__ import print_function

from ..subcommand import SubCommand


class Run(SubCommand):
    """Run the dataplicity service in the foreground"""

    help = """Run dataplicity agent"""

    def run(self):
        client = self.app.make_client()
        client.run_forever()



from __future__ import print_function


import logging
import os
import os.path
import time

from typing import Dict, Iterable, Iterator, List, Optional, Set, Tuple, TypedDict


try:
    # Scandir from stdlib in Python3
    from os import DirEntry, scandir
except ImportError:
    # Can't import scandir, so we will implement as much as we need
    # This will be less efficient, but will work

    from os.path import join
    from stat import S_ISDIR

    class DirEntry(object):  # type: ignore
        def __init__(self, path):
            # type(str) -> None
            self.path = path
            self.name = os.path.basename(self.path)
            self._stat = None

        def stat(self):
            # type() -> os.stat_result
            if self._stat is None:
                self._stat = os.stat(self.path)
            return self._stat

        def is_dir(self):
            # type: () -> bool
            return S_ISDIR(self.stat().st_mode)

        def is_file(self):
            # type: () -> bool
            return not S_ISDIR(self.stat().st_mode)

        def inode(self):
            # type: () -> int
            return self.stat().st_ino

    def scandir(dir_path):  # type: ignore
        # type(str) > Iterator[DirEntry]
        return iter(DirEntry(join(dir_path, path)) for path in os.listdir(dir_path))


FileInfo = Tuple[str, int]
DirectoryDict = TypedDict(
    "DirectoryDict", {"files": List[FileInfo], "dirs": List[str]}, total=False
)
ScanResult = TypedDict(
    "ScanResult",
    {"root": str, "time": float, "directories": Dict[str, DirectoryDict]},
)
_ScanStackEntry = Tuple[str, Iterable[DirEntry]]

log = logging.getLogger("agent")


class ScanDirectoryError(Exception):
    """Unable to scan the directory."""


def scan_directory(root_path, file_sizes=False, max_depth=10):
    # type: (str, bool, Optional[int]) -> ScanResult
    """Scan and serialize directory structure.

    Uses scandir to do this quite efficiently (without code recursion). Recursive links
    are detected and omitted.

    Returns a dict in the following format.

    {
        "root": <str: ABSOLUTE PATH>,
        "time": <float: EPOCH TIME>,
        "directories": {
            <str: RELATIVE PATH>: {
                "dirs" (optional): [<str: NAME>, ...],
                "files" (optional): [
                    (<str: NAME>, <int: FILESIZE>),
                    ...
                ]
            },
            ...
        }
    }

    Args:
        root_path (str): Root path
        file_sizes (boolean, optional): Add file sizes, defaults to False. If this is false, file sizes
            will be reported as -1
        max_depth (int, optional): Maximum number of depth of directory, defaults to 10

    Returns:
        dict: Serialized directory structure
    """
    root_path = os.path.abspath(os.path.expanduser(root_path))
    if not os.path.isdir(root_path):
        raise ScanDirectoryError("Can't scan %s; not a directory" % root_path)
    stack = []  # type: List[_ScanStackEntry]
    push = stack.append
    pop = stack.pop
    directories = {}  # type: Dict[str, DirectoryDict]
    visited_directories = set()  # type: Set[int]

    def push_directory(path):
        # type: (str) -> None
        """Push a new directory on to the stack."""
        scan_path = os.path.join(root_path, path)
        try:
            stack_entry = path, scandir(scan_path)
            push(stack_entry)
        except Exception as error:
            # Doesn't matter what the error is, log it and continue
            log.warning("error in scandir(%r); %s", scan_path, error)
        else:
            # This dict may contain "files" and "dirs", but are
            # omitted by default for brevity.
            directories[path] = {}

    scan_time = time.time()
    push_directory("")
    join = os.path.join
    while stack:
        path, scan = stack[-1]
        try:
            dir_entry = next(scan)
        except StopIteration:
            # End of directory scan, we can discard top of stack
            pop()
            continue
        if dir_entry.name.startswith("."):
            # Exclude hidden files and directories
            continue
        try:
            is_dir = dir_entry.is_dir()
        except Exception:
            is_dir = False

        try:
            is_file = dir_entry.is_file()
        except Exception:
            is_file = False

        if is_dir:
            if max_depth is not None and len(stack) >= max_depth:
                # Max depth reached, so skip this dir
                continue
            try:
                inode = dir_entry.inode()
            except Exception as error:
                log.warning("error in inode; %r", error)
            else:
                if inode in visited_directories:
                    # We have visited this directory before, we must have a recursive link
                    continue
                directories[path].setdefault("dirs", []).append(dir_entry.name)
                visited_directories.add(inode)
            push_directory(join(path, dir_entry.name))
        elif is_file:
            try:
                size = dir_entry.stat().st_size
            except Exception:
                size = -1
            file_info = (dir_entry.name, size) if file_sizes else (dir_entry.name, -1)
            directories[path].setdefault("files", []).append(file_info)

    scan_result = {
        "root": root_path,
        "time": scan_time,
        "directories": directories,
    }  # type: ScanResult
    return scan_result



from __future__ import print_function, unicode_literals

import json
import logging
import os.path
import tempfile
from threading import Lock, Thread
from time import time
from typing import Callable, Optional

from .compat import text_type
from .scan_directory import ScanDirectoryError, ScanResult, scan_directory

log = logging.getLogger("agent")


class DirectoryScanner(object):
    """Scans and serializes directory information."""

    def __init__(self, root_path):
        # type: (str) -> None
        """Create directory Scanner.

        Args:            
            root_path (str): Root of scan.            
        """
        self.root_path = root_path

        self._lock = Lock()

    def perform_scan(self, file_sizes=True, on_success=None):
        # type: (bool, Callable[[], None]) -> None
        """Perform a scan in the background."""
        if self._lock.locked():
            # Scan is in progress, no point in doing another
            return
        thread = Thread(
            target=self._perform_scan,
            kwargs={"file_sizes": file_sizes, "on_success": on_success},
        )
        thread.start()

    def _perform_scan(self, file_sizes=True, on_success=None):
        # type: (bool, Optional[Callable]) -> None
        """Scan and upload directory structure."""
        with self._lock:
            try:
                start_time = time()
                directory = scan_directory(self.root_path, file_sizes=file_sizes)
                elapsed = time() - start_time
                log.debug("scan directory elapsed %.1f", elapsed * 1000)
            except ScanDirectoryError as error:
                log.warning(str(error))
                raise
            except Exception:
                log.exception("failed to scan directory %s", self.root_path)
                raise
            else:
                try:
                    self.write_scan(directory)
                except Exception:
                    log.exception("failed to write_scan")
                    raise
                else:
                    if on_success is not None:
                        try:
                            on_success()
                        except Exception:
                            log.exception("error in on_success")

    def write_scan(self, directory):
        # type: (ScanResult) -> None
        """Save the scan to tmp."""

        scan_json = json.dumps(directory)
        if isinstance(scan_json, text_type):
            scan_json = scan_json.encode("utf-8")

        file_path = os.path.join(
            tempfile.gettempdir(), "__dataplicity_remote_directory_scan___.json"
        )
        with open(file_path, "wb") as scan_file:
            scan_file.write(scan_json)
